/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: DistributedJMXServerManager.java 16:34:49 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.communication.jmx;

import javax.management.MBeanServer;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.fractal.jmx.agent.AdminAttributes;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.tools.jmx.api.PetalsJMXConnection;
import org.ow2.petals.util.LoggingUtil;

/**
 * This class is a manager of JMX server. It is used to get a connection to a
 * JMX Server on a distant container.
 * 
 * @author ddesjardins - eBMWebsourcing
 * @author chamerling - eBMWebsourcing
 * @author Roland NAUDIN - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.communication.jmx.JMXService.class))
public class JMXServiceImpl implements JMXService {

    /**
     * Logger
     */
    protected LoggingUtil log;

    @Monolog(name = "logger")
    private Logger logger;

    /**
     * AdminAttributes
     */
    @Requires(name = "adminAtt", signature = org.objectweb.fractal.jmx.agent.AdminAttributes.class)
    protected AdminAttributes adminAttributes;

    @Requires(name = "topology", signature = org.ow2.petals.communication.topology.TopologyService.class, contingency = Contingency.OPTIONAL)
    protected TopologyService topologyService;

    /**
     * Create a {@link DistributedJMXServer} instance
     * 
     * @param containerName
     * @return
     * @throws PetalsException
     */
    public DistributedJMXServer createDistributedJMXServer(String containerName)
            throws PetalsException {
        log.start();
        DistributedJMXServer distributedJMXServer = null;

        if (topologyService == null) {
            // it is the stand alone distribution
            distributedJMXServer = new DistributedJMXServer(getLocalJMXServer());
        } else {
            try {
                ContainerConfiguration containerConfiguration = topologyService
                        .getContainerConfiguration(containerName);
                if (containerConfiguration != null) {
                    PetalsJMXConnection petalsJmxConnection = new PetalsJMXConnection(
                            containerConfiguration.getHost(), containerConfiguration
                                    .getJmxRMIConnectorPort(), containerConfiguration.getUser(),
                            containerConfiguration.getPassword());
                    distributedJMXServer = new DistributedJMXServer(petalsJmxConnection
                            .getMBeanServerConnection());
                } else {
                    throw new PetalsException("The configuration of the container " + containerName
                            + " can not be found. Can not connect to the JMX Server of container "
                            + containerName);
                }
            } catch (Exception e) {
                throw new PetalsException("Error while trying to connect to JMX server", e);
            }
        }
        log.end();
        return distributedJMXServer;
    }

    /**
     * Get the local JMX server
     * 
     * @return the local JMX server
     */
    public MBeanServer getLocalJMXServer() {
        return adminAttributes.getRawMBeanServer();
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
        log = new LoggingUtil(logger);
        log.call();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        log.call();
    }

}
