/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: RegistryUtil.java 10:30:18 AM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.communication.jndi.agent.util;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.Map.Entry;

import javax.naming.Binding;
import javax.naming.InvalidNameException;
import javax.naming.LinkRef;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameClassPair;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;
import javax.naming.NotContextException;

import org.ow2.petals.communication.jndi.agent.JNDIAgentServiceImpl;
import org.ow2.petals.communication.jndi.client.naming.JNDIConnection;
import org.ow2.petals.communication.jndi.client.naming.JNDILocalConnection;
import org.ow2.petals.communication.jndi.client.naming.NamingConnection;
import org.ow2.petals.communication.jndi.client.naming.NamingContextImpl;
import org.ow2.petals.util.LoggingUtil;

/**
 * Utility class to handle data.
 *
 * @author ddesjardins - eBMWebsourcing
 */
public class RegistryUtil {

    /**
     * The JNDI Agent implementation
     */
    private JNDIAgentServiceImpl jndiAgentServiceImpl;
    
    private LoggingUtil log;

    /**
     * Creates a new instance of {@link RegistryUtil}
     *
     * @param server
     */
    public RegistryUtil(JNDIAgentServiceImpl jndiAgentServiceImpl) {
        super();
        this.jndiAgentServiceImpl = jndiAgentServiceImpl;
        this.log = this.jndiAgentServiceImpl.getLog();
    }

    /**
     * Bind an object with a key in the specified context
     *
     * @param contextName
     *            name of the context
     * @param key
     *            key to bind with
     * @param value
     *            value to bind to the key
     * @throws NamingException
     *             if a problem occurs
     */
    public void bind(Object contextName, Object key, Object value)
        throws NamingException {
        log.start();

        if (!(key instanceof String)) {
            throw new InvalidNameException("Key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) key).length() == 0) {
            throw new InvalidNameException("Key must not be an empty String");
        }
        String path = "" + contextName + key;
        String fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        if (path.lastIndexOf("/") == 0) {
            fullContextName = (String) contextName;
        }

        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        if (jndiAgentServiceImpl.getData().containsKey(fullContextName)) {
            HashMap<String, Object> subData = jndiAgentServiceImpl.getData()
                .get(fullContextName);
            if (subData.containsKey(lastKey)) {
                throw new NameAlreadyBoundException(lastKey
                    + " is already bound in context " + fullContextName);
            } else {
                subData.put(lastKey, value);
                // update the replicated HashMap
                jndiAgentServiceImpl.getData().put(fullContextName, subData);
            }
        } else {
            throw new NotContextException("Context " + fullContextName
                + " does not exists");
        }

        log.end();
    }

    /**
     * Create and bind a new context in the specified context
     *
     * @param contextName
     *            name of the context in which we want to create a new context
     * @param key
     *            name of the context that we want to create
     * @return newly created context
     * @throws NamingException
     *             if the context name does not exist or if the key is already
     *             bound
     */
    public String createSubcontext(Object contextName, Object key)
        throws NamingException {
        log.start();

        if (!(key instanceof String)) {
            throw new InvalidNameException("Key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) key).length() == 0) {
            throw new InvalidNameException("Key must not be an empty String");
        }
        String path = "" + contextName + key;
        String fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        if (path.lastIndexOf("/") == 0) {
            fullContextName = (String) contextName;
        }
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        if (!jndiAgentServiceImpl.getData().containsKey(fullContextName)) {
            throw new NotContextException("Context " + fullContextName
                + " does not exists");
        } else {
            HashMap<String, Object> newMap = new HashMap<String, Object>();
            jndiAgentServiceImpl.getData().put(path + "/", newMap);
            bind(fullContextName, lastKey, newMap);
        }

        log.end();
        return "" + contextName + key + "/";
    }

    /**
     * Destroy a subcontext in the specified context
     *
     * @param contextName
     *            name of the context in which we want to create a new context
     * @param key
     *            name of the context that we want to create
     * @return newly created context
     * @throws NamingException
     *             if the context name does not exist or if the key is already
     *             bound
     */
    public void destroySubcontext(Object contextName, Object key)
        throws NamingException {
        log.start();

        if (!(key instanceof String)) {
            throw new InvalidNameException("Key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) key).length() == 0) {
            throw new InvalidNameException("Key must not be an empty String");
        }
        String path = "" + contextName + key;
        String fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        if (path.lastIndexOf("/") == 0) {
            fullContextName = (String) contextName;
        }
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        if (!jndiAgentServiceImpl.getData().containsKey(fullContextName)) {
            throw new NotContextException("Context " + fullContextName
                + " does not exists");
        } else {
            unbind(fullContextName, lastKey);
            jndiAgentServiceImpl.getData().remove(path + "/");
            for (String contextKey : jndiAgentServiceImpl.getData().keySet()) {
                if (contextKey.startsWith(path + "/")) {
                    jndiAgentServiceImpl.getData().remove(contextKey);
                }
            }
        }

        log.end();
    }

    /**
     * List the content of a context
     *
     * @param contextName
     *            name of the context
     * @param contextToList
     *            name of the context to list the content
     * @return array of NameClassPair
     * @throws NamingException
     */
    public NameClassPair[] list(Object contextName, Object contextToList)
        throws NamingException {
        log.start();

        if (!(contextToList instanceof String)) {
            throw new InvalidNameException(
                "The name of the context to list must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) contextToList).length() == 0) {
            throw new InvalidNameException("You should retrieve this context "
                + contextName + " by another way");
        }
        NameClassPair[] names = null;
        String path = (String) contextName + (String) contextToList + "/";
        if ("/".equals(contextToList)) {
            path = (String) contextToList;
        }

        if (!jndiAgentServiceImpl.getData().containsKey(path)) {
            throw new NotContextException("Context " + path
                + " does not exists");
        } else {
            Map<String, Object> mapToList = jndiAgentServiceImpl.getData().get(
                path);
            names = new NameClassPair[mapToList.keySet().size()];
            int i = 0;
            for (Entry<String, ?> entry : mapToList.entrySet()) {
                if (entry.getValue() != null) {
                    String className = entry.getValue().getClass().getName();
                    if (jndiAgentServiceImpl.getData().containsKey(
                        path + entry.getKey() + "/")) {
                        className = NamingContextImpl.class.getName();
                    }
                    names[i] = new NameClassPair(entry.getKey(), className);
                } else {
                    names[i] = new NameClassPair(entry.getKey(), null);
                }
                i++;
            }
        }
        log.end();
        return names;
    }

    /**
     * List the bindings of a context
     *
     * @param contextName
     *            name of the context
     * @param contextToList
     *            name of the context to list the content
     * @param localConnection
     *            {@code true} if the connection is local
     * @return array of NameClassPair
     * @throws NamingException
     */
    public Binding[] listBindings(Object contextName, Object contextToList,
        boolean localConnection) throws NamingException {
        log.start();
        if (!(contextToList instanceof String)) {
            throw new InvalidNameException(
                "The name of the context to list must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) contextToList).length() == 0) {
            throw new InvalidNameException("You should retrieve this context "
                + contextName + " by another way");
        }
        Binding[] bindings = null;
        Map<String, ?> mapToList = null;
        String path = "" + contextName + contextToList + "/";
        if ("/".equals(contextToList)) {
            path = (String) contextToList;
        }
        if (!jndiAgentServiceImpl.getData().containsKey(path)) {
            throw new NotContextException("Context " + path
                + " does not exists");
        } else {
            mapToList = jndiAgentServiceImpl.getData().get(path);
            bindings = new Binding[mapToList.keySet().size()];
            int i = 0;
            for (Entry<String, ?> entry : mapToList.entrySet()) {
                if (entry.getValue() != null) {
                    String className = entry.getValue().getClass().getName();
                    Object value = entry.getValue();
                    if (jndiAgentServiceImpl.getData().containsKey(
                        path + entry.getKey() + "/")) {
                        className = NamingContextImpl.class.getName();
                        NamingConnection namingConnection = null;
                        if (localConnection) {
                            namingConnection = new JNDILocalConnection(
                                jndiAgentServiceImpl);
                        } else {
                            namingConnection = new JNDIConnection(
                                jndiAgentServiceImpl.getDomainConfiguration()
                                    .getJndiPort(), jndiAgentServiceImpl
                                    .getDomainConfiguration().getJndiHost());
                        }
                        value = new NamingContextImpl(
                            new Hashtable<String, Object>(), namingConnection,
                            path + entry.getKey() + "/", localConnection);
                    }
                    bindings[i] = new Binding(entry.getKey(), className, value);
                } else {
                    bindings[i] = new Binding(entry.getKey(), null, null);
                }
                i++;
            }
        }
        log.end();
        return bindings;
    }

    /**
     * Lookup for an object
     *
     * @param contextName
     *            name of the context
     * @param key
     *            key to look for
     * @param localConnection
     *            {@code true} if the connection is local
     * @return value of the object bound to the key
     * @throws NamingException
     */
    public Object lookup(Object contextName, Object key, boolean localConnection)
        throws NamingException {
        log.start();

        if (!(key instanceof String)) {
            throw new InvalidNameException("Key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) key).length() == 0) {
            throw new InvalidNameException("You should retrieve this context "
                + contextName + " by another way");
        }
        Object out = null;
        String path = (String) contextName + (String) key;
        if (((String) key).startsWith("/")) {
            // The key is an absolute path (linkref)
            path = (String) key;
        } else if (((String) key).startsWith(".")) {
            // The key is a relative linkref
            path = contextName + ((String) key).substring(1);
        }
        String fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        if (path.lastIndexOf("/") == 0) {
            fullContextName = (String) contextName;
        }
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        if (!jndiAgentServiceImpl.getData().containsKey(fullContextName)) {
            throw new NotContextException("Context " + fullContextName
                + " does not exists");
        } else {
            if (jndiAgentServiceImpl.getData().get(fullContextName)
                .containsKey(lastKey)) {
                if (jndiAgentServiceImpl.getData().containsKey(
                    fullContextName + lastKey + "/")) {

                    NamingConnection namingConnection = null;
                    if (localConnection) {
                        namingConnection = new JNDILocalConnection(
                            jndiAgentServiceImpl);
                    } else {
                        namingConnection = new JNDIConnection(
                            jndiAgentServiceImpl.getDomainConfiguration()
                                .getJndiPort(), jndiAgentServiceImpl
                                .getDomainConfiguration().getJndiHost());
                    }
                    out = new NamingContextImpl(
                        new Hashtable<String, Object>(), namingConnection,
                        fullContextName + lastKey + "/", localConnection);
                } else {
                    out = jndiAgentServiceImpl.getData().get(fullContextName)
                        .get(lastKey);
                }
            } else {
                throw new NameNotFoundException(lastKey
                    + " is not bound in the context " + fullContextName);
            }
        }

        log.end();
        return out;
    }

    /**
     * Lookup for an object go through the links if the object found is a link
     *
     * @param contextName
     *            name of the context
     * @param key
     *            key to look for
     * @param localConnection
     *            {@code true} if the connection is local
     * @return value of the object bound to the key
     * @throws NamingException
     */
    public Object lookupLink(Object contextName, Object key,
        boolean localConnection) throws NamingException {
        log.start();

        if (!(key instanceof String)) {
            throw new InvalidNameException("Key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) key).length() == 0) {
            throw new InvalidNameException("You should retrieve this context "
                + contextName + " by another way");
        }
        Object out = null;
        boolean test = true;
        while (test) {
            out = lookup(contextName, key, localConnection);
            if (out instanceof LinkRef) {
                LinkRef linkRef = (LinkRef) out;
                contextName = "/";
                key = linkRef.getLinkName().substring(1);
            } else {
                test = false;
            }
        }

        log.end();
        return out;
    }

    /**
     * Rebind an object with a key in the specified context
     *
     * @param data
     *            data
     * @param contextName
     *            name of the context
     * @param key
     *            key to bind with
     * @param value
     *            value to bind to the key
     * @throws NamingException
     *             if a problem occurs
     */
    public void rebind(Object contextName, Object key, Object value)
        throws NamingException {
        log.start();

        if (!(key instanceof String)) {
            throw new InvalidNameException("Key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) key).length() == 0) {
            throw new InvalidNameException("Key must not be an empty String");
        }
        String path = "" + contextName + key;
        String fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        if (path.lastIndexOf("/") == 0) {
            fullContextName = (String) contextName;
        }
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        if (!jndiAgentServiceImpl.getData().containsKey(fullContextName)) {
            throw new NotContextException("Context " + fullContextName
                + " does not exists");
        } else {
//            if (jndiAgentServiceImpl.getData().get(fullContextName)
//                .containsKey(lastKey)) {
                HashMap<String, Object> subData = jndiAgentServiceImpl
                    .getData().get(fullContextName);
                subData.put(lastKey, value);
                // update the replicated HashMap
                jndiAgentServiceImpl.getData().put(fullContextName, subData);
//            } else {
//                throw new NameNotFoundException(lastKey
//                    + " is not bound in the context " + fullContextName);
//            }
        }
        log.end();
    }

    /**
     * Rebind an object with a key in the specified context
     *
     * @param contextName
     *            name of the context
     * @param oldKey
     *            old key
     * @param newKey
     *            new key to bind to
     * @param localConnection
     *            {@code true} if the connection is local
     * @throws NamingException
     *             if a problem occurs
     */
    public void rename(Object contextName, Object oldKey, Object newKey,
        boolean localConnection) throws NamingException {
        log.start();

        if (!(oldKey instanceof String)) {
            throw new InvalidNameException(
                "Old key must be non null and a String");
        }
        if (!(newKey instanceof String)) {
            throw new InvalidNameException(
                "New key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) oldKey).length() == 0) {
            throw new InvalidNameException(
                "Old key must not be an empty String");
        }
        if (((String) newKey).length() == 0) {
            throw new InvalidNameException(
                "New key must not be an empty String");
        }
        Object value = lookup(contextName, oldKey, localConnection);
        unbind(contextName, oldKey);
        bind(contextName, newKey, value);

        log.end();
    }

    /**
     * Unbind an an object
     *
     * @param contextName
     *            name of the context in which we work
     * @param key
     *            key to unbind
     * @throws NamingException
     */
    public void unbind(Object contextName, Object key) throws NamingException {
        log.start();

        if (!(key instanceof String)) {
            throw new InvalidNameException("Key must be non null and a String");
        }
        if (!(contextName instanceof String)) {
            throw new InvalidNameException(
                "Context name must be non null and a String");
        }
        if (((String) key).length() == 0) {
            throw new InvalidNameException("You should retrieve this context "
                + contextName + " by another way");
        }
        String path = "" + contextName + key;
        String fullContextName = path.substring(0, path.lastIndexOf("/") + 1);
        if (path.lastIndexOf("/") == 0) {
            fullContextName = (String) contextName;
        }
        String lastKey = path.substring(path.lastIndexOf("/") + 1);
        if (!jndiAgentServiceImpl.getData().containsKey(fullContextName)) {
            throw new NotContextException("Context " + fullContextName
                + " does not exists");
        } else {
            if (jndiAgentServiceImpl.getData().get(fullContextName)
                .containsKey(lastKey)) {
                HashMap<String, Object> subData = jndiAgentServiceImpl
                    .getData().get(fullContextName);
                subData.remove(lastKey);
                // update the replicated HashMap
                jndiAgentServiceImpl.getData().put(fullContextName, subData);
            } else {
                throw new NameNotFoundException(lastKey
                    + " is not bound in the context " + fullContextName);
            }
        }
        log.end();
    }
}
