/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.communication.network;

import static org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STANDALONE;

import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

import javax.naming.ConfigurationException;

import org.jgroups.Address;
import org.jgroups.Channel;
import org.jgroups.ChannelClosedException;
import org.jgroups.ChannelFactory;
import org.jgroups.ChannelNotConnectedException;
import org.jgroups.Global;
import org.jgroups.JChannelFactory;
import org.jgroups.Message;
import org.jgroups.Receiver;
import org.jgroups.View;
import org.jgroups.blocks.ReplicatedHashMap;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.network.data.JoinRequestData;
import org.ow2.petals.communication.network.data.JoinResponseData;
import org.ow2.petals.communication.network.data.NetworkData;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.util.LoggingUtil;

/**
 * This {@link NetworkService} implementation is based on JGROUPS.<br>
 * It uses 2 different stacks. One is based on UPD over multicast IP, the other
 * is based on TCP over unicast IP.<br>
 * The first stack is more efficient. For Networks whom multicast messages are
 * forbidden, the second stack must be used.
 * 
 * @author Roland NAUDIN - eBMWebSourcing
 * @since Petals 2.0
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.communication.network.NetworkService.class))
public class NetworkServiceImpl implements NetworkService, Receiver {

    private static final String STACKS_CONFIG_FILE = "stacks.xml";

    private static final String UDP_MCAST_STACK = "udp-sync";

    private static final String TCP_UCAST_STACK = "tcp-sync";

    /**
     * The latency max value. If no response is received over this value, the
     * responder is supposed absent
     */
    private static final int LATENTY_MAX_TIME = 5000;

    /**
     * The error message when the latency max value has been reached.
     */
    private static final String TIMEOUT_REACHED_ERROR = "The coordinator of the group do not respond";

    /**
     * Logging utilities
     */
    private LoggingUtil log;

    @Monolog(name = "logger")
    private Logger logger;

    /**
     * The topology service fractal component
     */
    @Requires(name = "topology", signature = org.ow2.petals.communication.topology.TopologyService.class)
    private TopologyService topologyService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * The domain configuration
     */
    private DomainConfiguration domainConfiguration;

    /**
     * The JGROUPS channel factory
     */
    private ChannelFactory channelFactory;

    /**
     * A channel for the Discovery service
     */
    private Channel discoveryCoordinationChannel;

    /**
     * A channel for the Discovery service
     */
    private Channel discoveryHashMapChannel;

    /**
     * The channel for the JNDI service
     */
    private Channel jndiHashMapChannel;

    /**
     * The replicated HashMap used by the JNDI service to store its initial
     * context data
     */
    private ReplicatedHashMap<String, HashMap<String, Object>> jndiHashMap;

    /**
     * The replicated HashMap container Addresses map
     */
    private ReplicatedHashMap<String, Address> containerAddressesHashMap;

    /**
     * Tricky Map used for standalone mode TODO: remove this map when the real
     * standalone mode would be implemented
     */
    private Map<String, HashMap<String, Object>> standaloneJndiHashMap;

    /**
     * The result of the joining request. Used to synchronized the response.
     */
    private String[] joinResult = new String[1];

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.communication.network.NetworkService#joinNetwork()
     */
    public String joinNetwork() throws NetworkException {
        this.log.start();

        synchronized (this.joinResult) {
            // always true if the node is the coordinator and so the first node
            if (this.discoveryCoordinationChannel.getView().getCreator().equals(
                    this.discoveryCoordinationChannel.getLocalAddress())) {
                this.joinResult[0] = JOIN_ACCEPTED;
            } else {
                this.joinResult[0] = TIMEOUT_REACHED_ERROR;
                NetworkData joinData = new JoinRequestData(this.containerConfiguration,
                        this.domainConfiguration);
                try {
                    this.discoveryCoordinationChannel.send(this.discoveryCoordinationChannel
                            .getView().getCreator(), null, joinData);
                    this.joinResult.wait(LATENTY_MAX_TIME);
                } catch (InterruptedException e) {
                    // do nothing
                } catch (ChannelNotConnectedException e) {
                    throw new NetworkException(e);
                } catch (ChannelClosedException e) {
                    throw new NetworkException(e);
                }
            }
        }

        this.log.end();
        return this.joinResult[0];
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.communication.network.NetworkService#leaveNetwork()
     */
    public void leaveNetwork() {
        this.log.start();

        if (this.jndiHashMapChannel != null && this.jndiHashMapChannel.isConnected()) {
            this.jndiHashMapChannel.close();
        }
        if (this.discoveryCoordinationChannel.isConnected()) {
            this.discoveryCoordinationChannel.close();
            this.discoveryHashMapChannel.close();
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.communication.network.NetworkService#getReplicatedHashMap()
     */
    public Map<String, HashMap<String, Object>> getReplicatedHashMap() {
        this.log.start();
        Map<String, HashMap<String, Object>> result = null;
        if (STANDALONE.equals(this.domainConfiguration.getMode())) {
            result = this.standaloneJndiHashMap;
        } else {
            result = this.jndiHashMap;
        }
        this.log.end();
        return result;
    }

    public void receive(Message message) {
        this.log.start();
        // Process the message only if this node is the destination
        if (message.getDest().equals(this.discoveryCoordinationChannel.getLocalAddress())) {
            NetworkData networkData = (NetworkData) message.getObject();

            if (networkData.getType().equals(NetworkData.MessageType.joinRequest)) {
                this.log
                        .info("A PEtALS container tries to join the Network, check its configuration");

                ContainerConfiguration newContainerConfiguration = ((JoinRequestData) message
                        .getObject()).getContainerConfiguration();
                DomainConfiguration newDomainConfiguration = ((JoinRequestData) message.getObject())
                        .getDomainConfiguration();
                String response = JOIN_ACCEPTED;
                try {
                    this.configurationService.verifyContainerTopology(newContainerConfiguration,
                            newDomainConfiguration);
                } catch (TopologyException e) {
                    this.log.warning("Refuse a PEtALS container named '"
                            + newContainerConfiguration.getName() + "' on "
                            + newContainerConfiguration.getHost() + " : " + e.getMessage());
                    response = e.getMessage();
                }

                if (JOIN_ACCEPTED.equals(response)) {
                    // refuse the candidate if the container is already
                    // registered as started
                    if (this.containerAddressesHashMap.containsKey(newContainerConfiguration
                            .getName())) {
                        response = "Container named '" + newContainerConfiguration.getName()
                                + "' is already started";
                        this.log.warning("Refuse a PEtALS container named '"
                                + newContainerConfiguration.getName() + "' on "
                                + newContainerConfiguration.getHost() + " : " + response);

                    } else {
                        // update the container addresses map
                        this.containerAddressesHashMap.put(newContainerConfiguration.getName(),
                                message.getSrc());
                        this.log.info("Accept the PEtALS container named '"
                                + newContainerConfiguration.getName() + "' on "
                                + newContainerConfiguration.getHost());
                    }
                }

                NetworkData responseData = new JoinResponseData(response);
                try {
                    this.discoveryCoordinationChannel.send(message.getSrc(), null, responseData);
                } catch (ChannelNotConnectedException e) {
                    this.log.error("Coordinator failed to send the join response",
                            new NetworkException(e));
                } catch (ChannelClosedException e) {
                    this.log.error("Coordinator failed to send the join response",
                            new NetworkException(e));
                }
            } else if (networkData.getType().equals(NetworkData.MessageType.joinResponse)) {
                synchronized (this.joinResult) {
                    this.joinResult[0] = ((JoinResponseData) message.getObject()).getResponse();
                    this.joinResult.notify();
                }
            }
        }

        this.log.end();
    }

    public void viewAccepted(View new_view) {
        this.log.start();

        // test if this node is the coordinator
        if (new_view.getMembers().get(0)
                .equals(this.discoveryCoordinationChannel.getLocalAddress())) {
            if (this.containerAddressesHashMap.size() == 0) {
                // add first element of the group in the map
                this.containerAddressesHashMap.put(this.containerConfiguration.getName(),
                        this.discoveryCoordinationChannel.getLocalAddress());
            } else if (this.containerAddressesHashMap.size() > new_view.getMembers().size()) {
                // remove the dead nodes and update their state in the topology
                for (String containerName : this.containerAddressesHashMap.keySet()) {
                    if (!new_view.getMembers().contains(
                            this.containerAddressesHashMap.get(containerName))) {
                        // update the topology
                        try {
                            this.topologyService.setContainerState(containerName,
                                    ContainerConfiguration.ContainerState.STOPPED);
                        } catch (PetalsException e) {
                            // in dynamic mode, the container may have leaved
                            // the domain (shut down)
                            if (!domainConfiguration.getMode().equals(
                                    DomainConfiguration.DomainMode.DYNAMIC)) {
                                this.log.error(
                                        "Failed to update the state to STOPPED of the container named '"
                                                + containerName + "'", e);
                            }
                        }
                        this.containerAddressesHashMap.remove(containerName);
                    }
                }
            }
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();
        this.domainConfiguration = this.configurationService.getDomainConfiguration();

        if (STANDALONE.equals(this.domainConfiguration.getMode())) {
            this.standaloneJndiHashMap = new Hashtable<String, HashMap<String, Object>>();
        } else {
            try {
                this.createChannels();

                this.startChannels();
            } catch (Exception e) {
                // log the error here as Fractal layer is removing the Exception
                // stack
                this.log.error("Failed to start Network service", e);
                throw e;
            }
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        this.log.call();

        if (!STANDALONE.equals(this.domainConfiguration.getMode())) {
            this.leaveNetwork();
        }
    }

    /**
     * Create the JGROUPS channels
     * 
     * @throws Exception
     */
    private void createChannels() throws Exception {
        this.log.start();

        System.setProperty(Global.BIND_ADDR, String.valueOf(this.containerConfiguration.getHost()));

        System.setProperty("jgroups.bind_port", String.valueOf(this.containerConfiguration
                .getNetworkBindingPort()));

        // Configure either the TCP_UCAST_STACK or the UDP_MCAST_STACK
        String stack;
        if (this.domainConfiguration.getNetworkUnicastAddressMap() != null) {
            stack = TCP_UCAST_STACK;
            String networkUnicastAdressList = "";
            Map<String, String> networkUnicastAddressMap = this.domainConfiguration
                    .getNetworkUnicastAddressMap();
            for (String networkUnicastAddress : networkUnicastAddressMap.values()) {
                networkUnicastAdressList = networkUnicastAdressList + networkUnicastAddress + ",";
            }
            networkUnicastAdressList.substring(0, networkUnicastAdressList.length() - 1);
            System.setProperty(Global.TCPPING_INITIAL_HOSTS, networkUnicastAdressList);
        } else if (this.domainConfiguration.getNetworkMuticastIp() != null) {
            stack = UDP_MCAST_STACK;
            System.setProperty(Global.UDP_MCAST_ADDR, this.domainConfiguration
                    .getNetworkMuticastIp());
            System.setProperty(Global.UDP_MCAST_PORT, String.valueOf(this.domainConfiguration
                    .getNetworkMuticastPort()));
        } else {
            throw new ConfigurationException("The network configuration is missing");
        }

        this.channelFactory = new JChannelFactory();
        URL stacksURL = this.getClass().getResource("/" + STACKS_CONFIG_FILE);
        if (stacksURL == null) {
            throw new IOException("Failed to reach the JGroups ressource '" + STACKS_CONFIG_FILE
                    + "'");
        }

        this.channelFactory.setMultiplexerConfig(stacksURL);

        this.discoveryCoordinationChannel = this.channelFactory
                .createMultiplexerChannel(stack, "A");
        this.discoveryCoordinationChannel.setReceiver(this);

        this.discoveryHashMapChannel = this.channelFactory.createMultiplexerChannel(stack, "B");
        this.containerAddressesHashMap = new ReplicatedHashMap<String, Address>(
                this.discoveryHashMapChannel, LATENTY_MAX_TIME);

        // create the JNDI Channel only if the PEtALS JNDI server
        // is used
        if (this.domainConfiguration.getJndiFactory() == null) {
            this.jndiHashMapChannel = this.channelFactory.createMultiplexerChannel(stack, "C");
            // TODO: turn the persistence on when it will works in JGROUPS
            // (defaults properties unsettable)
            this.jndiHashMap = new ReplicatedHashMap<String, HashMap<String, Object>>(
                    this.jndiHashMapChannel, false, LATENTY_MAX_TIME);
            // the updates on the replicated hashMap must be synchronous
            this.jndiHashMap.setBlockingUpdates(true);
        }

        this.log.end();
    }

    /**
     * Initialize the JGROUPS channel and connect them to the group.
     * 
     * @throws Exception
     */
    private void startChannels() throws Exception {
        this.log.start();

        // start the discovery HashMap and wait the synchronization of its data
        this.discoveryHashMapChannel.connect(this.domainConfiguration.getDomain());
        this.containerAddressesHashMap.start(LATENTY_MAX_TIME);

        this.discoveryCoordinationChannel.connect(this.domainConfiguration.getDomain());

        if (this.jndiHashMapChannel != null) {
            // start the JNDI HashMap and wait the synchronization of its data
            this.jndiHashMapChannel.connect(this.domainConfiguration.getDomain());
            this.jndiHashMap.start(LATENTY_MAX_TIME);
        }

        this.log.end();
    }

    public byte[] getState() {
        // do nothing
        return null;
    }

    public void setState(byte[] arg0) {
        // do nothing

    }

    public void block() {
        // do nothing

    }

    public void suspect(Address arg0) {
        // do nothing

    }
}
