
package org.ow2.petals.jbi.management.admin;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.objectweb.fractal.adl.ADLException;
import org.ow2.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.container.lifecycle.ServiceAssemblyLifeCycle;
import org.ow2.petals.container.lifecycle.SharedLibraryLifeCycle;
import org.ow2.petals.jbi.management.MBeanNames;
import org.ow2.petals.jbi.management.MBeanNamesImpl;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.kernel.server.MBeanHelper;
import org.ow2.petals.util.LoggingUtil;

/**
 * <code>LifeCycleManagerImpl</code> is the main class for JMX management. It
 * is the entry point to register and unregister elements in the JMX environment
 * (elements are JBI binding/engine components, service assemblies, installers
 * and systemServices components). Related MBeans are created when registering
 * and unregistering elements. <code>LifeCycleManagerImpl</code> owns the JMX
 * management services offered by the JBI container to administrative clients.
 * Management services are : - AdminServiceImpl (referenced in the JMX server as
 * "admin"); - InstallationServiceMBean (referenced in the JMX server as
 * "installation"); - DeploymentServiceMBean (referenced in the JMX server as
 * "deployment"); <code>LifeCycleManagerImpl</code> is the central point for
 * accessing the mBeans associated to the JBI container elements.
 * <code>ManagementException</code> exceptions are thrown when problems for
 * accessing the JMX server occur. TODO refactor to get System Services from
 * Fractal, and BC and SE from JMX
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
public class LifeCycleManagerImpl {
    protected Map<ObjectName, ComponentLifeCycle> bindingCompoLifeCycles;

    protected Map<ObjectName, ComponentLifeCycle> engineCompoLifeCycles;

    protected Map<String, ServiceAssemblyLifeCycle> serviceAssemblyLifeCycles;

    protected Map<String, SharedLibraryLifeCycle> sharedLibraryLifeCycles;

    protected Map<ObjectName, Installer> installers;

    protected MBeanNames mBeanNames;

    protected MBeanServer mbeanServer;

    /**
     * Logger wrapper
     */
    private LoggingUtil log;

    /**
     * Default constructor. Creates a <code>Manager</code> instance. JMX
     * server is created.
     * 
     */
    public LifeCycleManagerImpl(MBeanServer mbeanServer, LoggingUtil log) {

        this.log = log;

        this.mbeanServer = mbeanServer;

        this.mBeanNames = new MBeanNamesImpl("org.ow2.petals");

        this.bindingCompoLifeCycles = Collections
                .synchronizedMap(new HashMap<ObjectName, ComponentLifeCycle>());

        this.engineCompoLifeCycles = Collections
                .synchronizedMap(new HashMap<ObjectName, ComponentLifeCycle>());

        this.serviceAssemblyLifeCycles = Collections
                .synchronizedMap(new HashMap<String, ServiceAssemblyLifeCycle>());

        this.installers = Collections.synchronizedMap(new HashMap<ObjectName, Installer>());

        this.sharedLibraryLifeCycles = Collections
                .synchronizedMap(new HashMap<String, SharedLibraryLifeCycle>());

    }

    // /////////////////////////
    // /// retrieve platform component MBeanNames
    // /////////////////////////
    /**
     * Retrieve the AdminServieMBean ObjectName.
     */
    public ObjectName getAdminServiceMBeanName() throws ManagementException {
        this.log.start();

        ObjectName result = null;

        try {
            result = MBeanHelper.getMBeanHelper().retrieveServiceMBean(MBeanHelper.ADMIN_MBEAN,
                    this.mbeanServer);
        } catch (MalformedObjectNameException e) {
            throw new ManagementException("Failed to retrieve the MBean name of the Admin service",
                    e);
        } catch (ADLException e) {
            throw new ManagementException("Failed to retrieve the MBean name of the Admin service",
                    e);
        }
        if (result == null) {
            throw new ManagementException(
                    "Failed to retrieve the MBean name of the Admin service. The name '"
                            + MBeanHelper.ADMIN_MBEAN + "' doesn't exist");
        }

        this.log.end();
        return result;
    }

    /**
     * Retrieve the DeploymentServiceMBean ObjectName. This is used by a
     * ServiceAssembly LifeCycle. As there is no MBean for the SA LifeCycle, the
     * DeploymentServiceMbean is used to represent it. All SA notifications are
     * reported to the DeploymentLifeCycleMB
     * 
     * @return the jmx name of the deploymentServiceMBean
     */
    public ObjectName getDeploymentServiceMBeanName() throws ManagementException {
        this.log.start();

        ObjectName result = null;

        try {
            result = MBeanHelper.getMBeanHelper().retrieveServiceMBean(
                    MBeanHelper.DEPLOYMENT_MBEAN, this.mbeanServer);

        } catch (MalformedObjectNameException e) {
            throw new ManagementException(
                    "Failed to retrieve the MBean name of the Deployment service", e);
        } catch (ADLException e) {
            throw new ManagementException(
                    "Failed to retrieve the MBean name of the Deployment service", e);
        }
        if (result == null) {
            throw new ManagementException(
                    "Failed to retrieve the MBean name of the Deployment service. The name '"
                            + MBeanHelper.ADMIN_MBEAN + "' doesn't exist");
        }

        this.log.end();
        return result;
    }

    /**
     * Retrieve the InstallationServiceMBean ObjectName.
     */
    public ObjectName getInstallationServiceMBeanName() throws ManagementException {
        this.log.start();

        ObjectName result = null;
        try {
            result = MBeanHelper.getMBeanHelper().retrieveServiceMBean(
                    MBeanHelper.INSTALLATION_MBEAN, this.mbeanServer);
        } catch (MalformedObjectNameException e) {
            throw new ManagementException(
                    "Failed to retrieve the MBean name of the Installation service", e);
        } catch (ADLException e) {
            throw new ManagementException(
                    "Failed to retrieve the MBean name of the Installation service", e);
        }
        if (result == null) {
            throw new ManagementException(
                    "Failed to retrieve the MBean name of the Installation service. The name '"
                            + MBeanHelper.ADMIN_MBEAN + "' doesn't exist");
        }

        this.log.end();
        return result;
    }

    // ////////////////////////////
    // manage components : bindings and engines
    // ////////////////////////////

    /**
     * Register a mbean in the JMX server representing the specified binding
     * component.
     * 
     * @param componentElement
     * @throws ManagementException
     */
    public ObjectName registerBindingComponent(ComponentLifeCycle lifeCycle)
            throws ManagementException {
        this.log.start();

        ObjectName name = lifeCycle.getMBeanName();

        synchronized (this.bindingCompoLifeCycles) {
            try {
                this.mbeanServer.registerMBean(lifeCycle, name);
                this.bindingCompoLifeCycles.put(name, lifeCycle);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }

        this.log.end();
        return name;
    }

    /**
     * Register a mbean in the JMX server representing the specified engine
     * component
     * 
     * @param componentElement
     * @throws ManagementException
     */
    public ObjectName registerEngineComponent(ComponentLifeCycle lifeCycle)
            throws ManagementException {
        this.log.start();

        ObjectName name = lifeCycle.getMBeanName();

        synchronized (this.engineCompoLifeCycles) {
            try {
                this.mbeanServer.registerMBean(lifeCycle, name);
                this.engineCompoLifeCycles.put(name, lifeCycle);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }

        this.log.end();
        return name;
    }

    /**
     * @see org.ow2.petals.jbi.management.admin.LifeCycleManagerService#getComponentByName(java.lang.String)
     */
    public ComponentLifeCycle getComponentLifeCycleByName(String name) {
        this.log.call();

        ComponentLifeCycle on = null;

        // search in bindings
        on = this.getBindingComponentByName(name);
        if (on == null) {
            // search in engines
            on = this.getEngineComponentByName(name);
        }

        return on;
    }

    /**
     * Unregister the binding component of the JMX server
     * 
     * @throws ManagementException
     */
    public void unregisterBindingComponent(ObjectName name) throws ManagementException {
        this.log.start();

        synchronized (this.bindingCompoLifeCycles) {
            try {
                this.mbeanServer.unregisterMBean(name);
                this.bindingCompoLifeCycles.remove(name);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }

        this.log.end();
    }

    /**
     * Unregister the engine component of the JMX server
     * 
     * @throws ManagementException
     */
    public void unregisterEngineComponent(ObjectName name) throws ManagementException {
        this.log.start();

        synchronized (this.engineCompoLifeCycles) {
            try {
                this.mbeanServer.unregisterMBean(name);
                this.engineCompoLifeCycles.remove(name);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }

        this.log.end();
    }

    // ////////////////////////////
    // manage installers
    // ////////////////////////////
    /**
     * Register an installer mbean in the JMX server representing the specified
     * component installer.
     * 
     * @param installerMBean
     * @throws ManagementException
     */
    public ObjectName registerInstaller(Installer installerMBean) throws ManagementException {
        this.log.start();

        ObjectName installerMBeanName = installerMBean.getMBeanName();

        synchronized (this.installers) {
            try {
                this.mbeanServer.registerMBean(installerMBean, installerMBeanName);
                this.installers.put(installerMBeanName, installerMBean);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }

        this.log.end();
        return installerMBeanName;
    }

    /**
     * Return the component installer associated with this component name.
     * 
     * @param name
     *            the component name
     * @return The related installer, or null
     */
    @SuppressWarnings("unchecked")
    public ObjectName getInstallerByName(String name) {
        this.log.call();

        ObjectName result = null;

        for (ObjectName objectName : this.installers.keySet()) {
            if (this.installers.get(objectName).getComponentName().equals(name)) {
                result = objectName;
                break;
            }
        }

        return result;
    }

    /**
     * Return the component installer MBean associated with this componant name.
     * 
     * @param name
     *            the installer Object name
     * @return The related installer, or null
     */
    public Installer getInstallerMBeanByName(ObjectName name) {
        return this.installers.get(name);
    }

    /**
     * Unregister the component installer of the JMX server
     * 
     * @throws ManagementException
     */
    public void unregisterInstaller(ObjectName name) throws ManagementException {
        this.log.start();

        synchronized (this.installers) {
            try {
                this.mbeanServer.unregisterMBean(name);
                this.installers.remove(name);
            } catch (Exception e) {
                throw new ManagementException(e);
            }
        }

        this.log.end();

    }

    // ////////////////////////////
    // manage service assemblies
    // ////////////////////////////
    /**
     * Register a ServiceAssemblyLifeCycle into the a synchronized Map<String,ServiceAssemblyLifeCycle>
     * 
     * @param saName
     * @param serviceAssemblyLifeCycle
     */
    public void registerServiceAssembly(String saName,
            ServiceAssemblyLifeCycle serviceAssemblyLifeCycle) {
        this.log.call();

        this.serviceAssemblyLifeCycles.put(saName, serviceAssemblyLifeCycle);
    }

    /**
     * Return the service assembly associated with this name.
     * 
     * @param saName
     * @return The related ServiceAssemblyLifeCycle, or null
     */
    public ServiceAssemblyLifeCycle getServiceAssemblyByName(String saName) {
        return this.serviceAssemblyLifeCycles.get(saName);
    }

    /**
     * Unregister the ServiceAssembly life cycle from the sync map
     * 
     * @throws ManagementException
     */
    public void unregisterServiceAssembly(String saName) throws ManagementException {
        this.log.call();

        this.serviceAssemblyLifeCycles.remove(saName);
    }

    // ////////////////////////////
    // manage shared library
    // ////////////////////////////
    /**
     * Register a SharedLibraryLifeCycle into the a synchronized Map<String,SharedLibraryLifeCycle>
     * 
     * @param slName
     * @param sharedLibraryLifeCycle
     */
    public void registerSharedLibrary(String slName, SharedLibraryLifeCycle sharedLibraryLifeCycle) {
        this.log.call();

        this.sharedLibraryLifeCycles.put(slName, sharedLibraryLifeCycle);

    }

    /**
     * Return the shared library associated with this name.
     * 
     * @param slName
     * @return The related SharedLibraryLifeCycle, or null
     */
    public SharedLibraryLifeCycle getSharedLibraryByName(String slName) {
        return this.sharedLibraryLifeCycles.get(slName);
    }

    /**
     * Unregister the Shared library life cycle from the sync map
     * 
     * @throws ManagementException
     */
    public void unregisterSharedLibrary(String slName) throws ManagementException {
        this.log.call();

        this.sharedLibraryLifeCycles.remove(slName);
    }

    // //////////////////////////////
    // / getters and setters
    // //////////////////////////////

    /**
     * Return the <code>MBeanNamesImpl</code> element which allows the
     * creation of mBeans names.
     * 
     * @return
     */
    public MBeanNames getMBeanNames() {
        return this.mBeanNames;
    }

    /**
     * Return the JMX server used by the JBI container.
     * 
     * @return
     */
    public MBeanServer getMBeanServer() {
        return this.mbeanServer;
    }

    /**
     * Return the synchronized map of binding components.
     * 
     * @return
     */
    public Map<ObjectName, ComponentLifeCycle> getBindingCompoLifeCycles() {
        return this.bindingCompoLifeCycles;
    }

    /**
     * Return the synchronized map of engine components.
     * 
     * @return
     */
    public Map<ObjectName, ComponentLifeCycle> getEngineCompoLifeCycles() {
        return this.engineCompoLifeCycles;
    }

    /**
     * Return the synchronized map of installers.
     * 
     * @return
     */
    public Map<ObjectName, Installer> getInstallers() {
        return this.installers;
    }

    /**
     * Return the synchronized map of deployed service assembly.
     * 
     * @return
     */
    public Map<String, ServiceAssemblyLifeCycle> getServiceAssemblyLifeCycles() {
        return this.serviceAssemblyLifeCycles;
    }

    /**
     * Return the synchronized map of deployed shared library.
     * 
     * @return
     */
    public Map<String, SharedLibraryLifeCycle> getSharedLibraryLifeCycles() {
        return this.sharedLibraryLifeCycles;
    }
    

    /**
     * Return the binding component associated with this name.
     * 
     * @param name
     * @return The related objectName, or null
     */
    private ComponentLifeCycle getBindingComponentByName(String name) {
        this.log.call();

        ComponentLifeCycle result = null;

        for (ComponentLifeCycle lifeCycle : this.bindingCompoLifeCycles.values()) {
            if (lifeCycle.getName().equals(name)) {
                result = lifeCycle;
                break;
            }
        }

        return result;
    }

    /**
     * Return the engine component associated with this name.
     * 
     * @param name
     * @return The related objectName, or null
     */
    private ComponentLifeCycle getEngineComponentByName(String name) {
        this.log.call();

        ComponentLifeCycle result = null;

        for (ComponentLifeCycle lifeCycle : this.engineCompoLifeCycles.values()) {
            if (lifeCycle.getName().equals(name)) {
                result = lifeCycle;
                break;
            }
        }

        return result;
    }

}
