/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: EndpointRegistry.java 11:27:27 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.registry;

import static org.ow2.petals.Constants.UNRESOLVED_INTERFACE;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.jmx.DistributedJMXServer;
import org.ow2.petals.communication.jmx.JMXService;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.LinkedEndpoint;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ParameterCheckHelper;
import org.ow2.petals.util.WSDLUtil;
import org.ow2.petals.util.XMLUtil;
import org.w3c.dom.Document;

/**
 * This class is used to manage everything that deals with endpoint in the
 * distributed environment
 * 
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.messaging.registry.EndpointRegistry.class))
public class EndpointRegistryImpl implements EndpointRegistry {

    private static final String UNDEFINED = "undefined";

    private static final String KEY_TYPE = "type";

    private static final String KEY_INTERFACE_NAMES = "interfaceNames";

    private static final String KEY_SERVICE_NAME = "serviceName";

    private static final String KEY_ENDPOINT_NAME = "endpointName";

    private static final String KEY_CONTAINER_NAME = "containerName";

    private static final String KEY_COMPONENT_NAME = "componentName";

    private static final String TYPE_INTERNAL = "internal";

    private static final String TYPE_EXTERNAL = "external";

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Registry
     */
    protected JNDIRegistry registry;

    /**
     * Manager of the distributed JMX server
     */
    @Requires(name = "jmx", signature = org.ow2.petals.communication.jmx.JMXService.class)
    protected JMXService serverManager;

    /**
     * JNDI service fractal component
     */
    @Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    protected ConfigurationService configurationService;

    /**
     * @see EndpointRegistry#activateEndpoint(QName, String, ConsumerEndpoint)
     */
    public InternalEndpoint activateEndpoint(QName serviceName, String endpointName,
            ConsumerEndpoint address) throws JBIException {
        this.log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        if (address == null) {
            throw new IllegalArgumentException("Consumer endpoint must be non null");
        }
        InternalEndpoint se = new InternalEndpoint(serviceName, endpointName, address
                .getComponentName(), address.getContainerName(), address.getSubDomainName(), this);
        this.registry.registerInternalEndpoint(se);
        this.log.end();
        return se;
    }

    /**
     * @see EndpointRegistry#createConnection(QName, QName, String)
     */
    public void createConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException("Interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.registry.registerConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#createConnection(QName, String, QName, String)
     */
    public void createConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.start();
        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        this.registry.registerConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#deactivateEndpoint(ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.unregisterInternalEndpoint((AbstractEndpoint) endpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#deleteConnection(QName, QName, String)
     */
    public void deleteConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException("Consumer interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.registry.unregisterConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#deleteConnection(QName, String, QName, String)
     */
    public void deleteConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.start();
        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        this.registry.unregisterConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#deregisterExternalEndpoint(ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.unregisterExternalEndpoint(endpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#getEndpoint(QName, String)
     * 
     * <ol>
     * <li>Retreive the named endpoint</li>
     * <li>If found, update the endpoint service of the found endpoint with
     * this</li>
     * </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        this.log.call();
        return this.getEndpoint(service, name, true);
    }

    /**
     * @see EndpointRegistry#getEndpoint(QName, String,boolean)
     * 
     * <ol>
     * <li>Retrieve the named endpoint</li>
     * <li>If found, update the endpoint service of the found endpoint with
     * this</li>
     * </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name, boolean resolveLink) {
        this.log.start();
        if (service == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (name == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        AbstractEndpoint abstractEndpoint = null;
        try {
            abstractEndpoint = this.registry.getInternalEndpoint(service, name, resolveLink);
        } catch (RegistryException e) {
            // no endpoint found, return null;
        }
        if (abstractEndpoint != null && abstractEndpoint.getEndpointRegistry() == null) {
            abstractEndpoint.setEndpointRegistry(this);
        }
        this.log.end();
        return abstractEndpoint;
    }

    /**
     * @see EndpointRegistry#getEndpointDescriptorForEndpoint(ServiceEndpoint)
     */
    public Document getEndpointDescriptorForEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        Document document = null;
        try {

            // Update the new endpoints
            this.updateNewEndpoints();

            // Recover the endpoint
            AbstractEndpoint abstractEndpoint = this.registry.getInternalEndpoint(endpoint
                    .getServiceName(), endpoint.getEndpointName());

            document = abstractEndpoint.getEndpointDescriptor();
        } catch (RegistryException e) {
            this.log.error("Error while validating endpoint " + endpoint.getEndpointName(), e);
        }
        this.log.end();
        return document;
    }

    /**
     * @see EndpointRegistry#getExternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForInterface(QName interfaceName) {
        this.log.start();
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            endpoints = this.registry.getExternalEndpointsForInterface(interfaceName);
            // Update the EndpointRegistry field of the found endpoints
            for (int i = 0; i < endpoints.length; i++) {
                AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
                if (abstractEndpoint.getEndpointRegistry() == null) {
                    abstractEndpoint.setEndpointRegistry(this);
                    endpoints[i] = abstractEndpoint;
                }
            }
        } catch (RegistryException e) {
            this.log.error("Error while getting external endpoints for interface " + interfaceName, e);
        }
        this.log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#getExternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        this.log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = null;
        endpoints = this.registry.getExternalEndpointsForService(serviceName);
        // Update the EndpointRegistry field of the found endpoints
        for (int i = 0; i < endpoints.length; i++) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
            if (abstractEndpoint.getEndpointRegistry() == null) {
                abstractEndpoint.setEndpointRegistry(this);
                endpoints[i] = abstractEndpoint;
            }
        }
        this.log.end();
        return endpoints;
    }

    /**
     * Return the interface for the specified endpoint
     * 
     * @param endpoint
     *            endpoint on which we want the interface names; must be non
     *            null
     * @return array of interfaces QName
     */
    public QName[] getInterfacesForEndpoint(AbstractEndpoint endpoint) {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        QName[] interfs = new QName[0];
        AbstractEndpoint abstractEndpoint = null;
        try {
            this.updateNewEndpoints();
            abstractEndpoint = this.registry.getInternalEndpoint(endpoint.getServiceName(), endpoint
                    .getEndpointName());
            interfs = abstractEndpoint.getInterfaces();
        } catch (Exception e) {
            this.log.error("Error while getting interfaces for endpoint " + endpoint.getEndpointName(),
                    e);
        }
        this.log.end();
        return interfs;
    }

    /**
     * @see EndpointRegistry#getInternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForInterface(QName interfaceName) {
        this.log.start();
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            // Update the new endpoints
            this.updateNewEndpoints();

            endpoints = this.registry.getInternalEndpointsForInterface(interfaceName);
            for (int i = 0; i < endpoints.length; i++) {
                AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
                if (abstractEndpoint.getEndpointRegistry() == null) {
                    abstractEndpoint.setEndpointRegistry(this);
                    endpoints[i] = abstractEndpoint;
                }
            }
        } catch (RegistryException e) {
            this.log.error("Error while getting internal endpoints for interface " + interfaceName, e);
        } catch (JBIException e) {
            this.log.error("Error while getting internal endpoints for interface " + interfaceName, e);
        }
        this.log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#getInternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForService(QName serviceName) {
        this.log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        endpoints = this.registry.getInternalEndpointsForService(serviceName);
        for (int i = 0; i < endpoints.length; i++) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
            if (abstractEndpoint.getEndpointRegistry() == null) {
                abstractEndpoint.setEndpointRegistry(this);
                endpoints[i] = abstractEndpoint;
            }
        }
        this.log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#isExchangeWithConsumerOkayForComponent(InternalEndpoint,
     *      MessageExchange)
     */
    public boolean isExchangeWithConsumerOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        ParameterCheckHelper.isNullParameterWithLog(internalEndpoint, "Endpoint", this.log);
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = this.serverManager.createDistributedJMXServer(internalEndpoint
                    .getContainerName());
            // Ask the component if it is ok to accept the exchange
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_CONS, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            this.log.error("Problem in isExchangeWithConsumerOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /**
     * @see EndpointRegistry#isExchangeWithProviderOkayForComponent(InternalEndpoint,
     *      MessageExchange)
     */
    public boolean isExchangeWithProviderOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        if (internalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = this.serverManager.createDistributedJMXServer(internalEndpoint
                    .getContainerName());
            // Retreive the endpoint description
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_PROV, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            this.log.error("Problem in isExchangeWithProviderOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /**
     * @see EndpointRegistry#registerExternalEndpoint(ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint) throws JBIException {
        this.log.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.registerExternalEndpoint(externalEndpoint);
        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
        this.log = new LoggingUtil(this.logger);
        this.log.call();
        this.registry = new JNDIRegistry(this.configurationService, this.jndiService, this.log);
        try {
            this.registry.initContexts();
        } catch (RegistryException e) {
            throw new IllegalLifeCycleException(e.toString());
        }

    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        this.log.call();
        this.registry = null;
    }

    public Document getDescription(String serviceName, String endpointName) throws Exception {
        ServiceEndpoint serviceEndpoint = this.getEndpoint(QName.valueOf(serviceName), endpointName);
        try {
            return this.getEndpointDescriptorForEndpoint(serviceEndpoint);
        } catch (JBIException e) {
            String msg = "Failed to retrieve endpoint description for : servicename=" + serviceName
                    + " and endpointname=" + endpointName;
            this.log.error(msg, e);
            throw new Exception(msg);
        }
    }

    public List<Map<String, Object>> getAllExternalEndpoints() {
        List<Map<String, Object>> externalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] externalEndpoints = this.getExternalEndpointsForInterface(null);
        for (ServiceEndpoint serviceEndpoint : externalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName());
            List<String> interfaceNames = this.interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            endpointMap.put(KEY_TYPE, TYPE_EXTERNAL);
            externalEndpointsMap.add(endpointMap);
        }
        return externalEndpointsMap;
    }

    public List<Map<String, Object>> getAllInternalEndpoints() {
        List<Map<String, Object>> internalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] internalEndpoints = this.getInternalEndpointsForInterface(null);
        for (ServiceEndpoint serviceEndpoint : internalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName().toString());
            List<String> interfaceNames = this.interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            if (serviceEndpoint instanceof LinkedEndpoint) {
                LinkedEndpoint linkEP = (LinkedEndpoint) serviceEndpoint;
                endpointMap.put(KEY_TYPE, linkEP.getToServiceName().toString() + "@"
                        + linkEP.getToEndpointName());
            } else {
                endpointMap.put(KEY_TYPE, TYPE_INTERNAL);
            }
            internalEndpointsMap.add(endpointMap);
        }
        return internalEndpointsMap;
    }

    public List<Map<String, Object>> getAllEndpoints() {
        List<Map<String, Object>> allEndpointsMap = this.getAllExternalEndpoints();
        allEndpointsMap.addAll(this.getAllInternalEndpoints());
        return allEndpointsMap;
    }

    /**
     * Update all the new endpoints and remove them. The method is synchronized
     * to avoid reentrant problems.
     * 
     * @throws JBIException
     *             if a problem occurs
     */
    protected synchronized void updateNewEndpoints() throws JBIException {

        for (AbstractEndpoint abstractEndpoint : this.registry.retrieveNewEndpoints()) {
            this.updateEndpointDescriptionAndInterfaces(abstractEndpoint);
            try {
                this.registry.validateEndpoint(abstractEndpoint);
            } catch (JBIException e) {
                this.log.error("Problem while updating new endpoint '"
                        + abstractEndpoint.getEndpointName() + "'", e);
            }
        }
        // We clean the new endpoints context
        this.registry.cleanNewEndpoints();

    }

    /**
     * This method will update the given endpoint description and its
     * interfaces.
     * 
     * @param endpoint
     *            endpoint to update
     */
    private void updateEndpointDescriptionAndInterfaces(final AbstractEndpoint endpoint) {
        this.log.start();

        String description = null;
        List<QName> interfaces = new ArrayList<QName>();

        try {
            // Get a connection to the JMX server of the container were the
            // endpoint is deployed
            DistributedJMXServer server = this.serverManager.createDistributedJMXServer(endpoint
                    .getContainerName());
            // Retrieve the endpoint description
            description = (String) server.invoke(server.getAdminServiceMBeanName(),
                    DistributedJMXServer.GET_SERVICE_DESCRITION, new Object[] {
                            endpoint.getComponentName(), endpoint }, new String[] {
                            "java.lang.String", "javax.jbi.servicedesc.ServiceEndpoint" });

        } catch (Exception e) {
            this.log.warning("Failed to retrieve the description from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName() + "'" + "'. Set its interface to "
                    + UNRESOLVED_INTERFACE, e);
            interfaces.add(UNRESOLVED_INTERFACE);
        }

        try {
            // Get the whole description
            if (description == null) {
                throw new PetalsException("The description retrieved for the endpoint '"
                        + endpoint.getEndpointName() + "' is null");
            } else {
                endpoint.setDescription(description);

                // Get the interfaces
                interfaces = WSDLUtil.getInterfacesForService(XMLUtil
                        .createDocumentFromString(description), endpoint.getServiceName());
            }
        } catch (PetalsException e) {
            this.log.warning("Failed to retrieve interfaces from the description of the endpoint '"
                    + endpoint.getEndpointName() + "'. Set its interface to "
                    + UNRESOLVED_INTERFACE, e);
            interfaces.add(UNRESOLVED_INTERFACE);
        }

        endpoint.setInterfaces(interfaces.toArray(new QName[0]));

        this.log.end();
    }

    private List<String> interfaceQNameToString(AbstractEndpoint abstractEndpoint) {
        QName[] interfaces = abstractEndpoint.getInterfaces();
        List<String> interfaceNames = new ArrayList<String>();
        if (interfaces != null && interfaces.length > 0) {
            for (QName interfaceName : interfaces) {
                interfaceNames.add(interfaceName.toString());
            }
        } else {
            interfaceNames.add(UNDEFINED);
        }
        return interfaceNames;
    }

}
