/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramRegistry.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.registry;

import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.LinkRef;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NotContextException;
import javax.xml.namespace.QName;

import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ExternalEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.LinkedEndpoint;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.util.JNDIUtil;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.NamingHelper;
import org.ow2.petals.util.StringHelper;

/**
 * This registry is the registry used in a distributed environment. It used a
 * JNDI distributed directory to manage the endpoint storage. <br>
 * Only subdomain-endpoints or public endpoints from other subdomains are
 * retrieved by the getXXX() methods
 * 
 * This is the JNDIRegistry service format :
 * 
 * <pre>
 *  '/'
 *  '/endpoints/'
 *  '          /e1' (key is serviceName@endpointName)
 *  '          /e2'
 *  '/services/'
 *  '         /s1/ref-e1' (key is serviceName and object is link to /endpoints/e1)
 *  '         /s2/ref-e2'
 *  '/interfaces/'
 *  '           /i1/ref-e1' (key is itfName and object is link to /endpoints/e1)
 *  '           /i2/ref-e2'
 *  '/new-endpoints/'
 *  '              /ref-e1' (link to /endpoints/e1)
 *  '              /ref-e2' (link to /endpoints/e2)
 * </pre>
 * 
 * @author Adrien LOUIS
 * @author ddesjardins - eBMWebsourcing
 * @author chamerling - eBM WebSourcing
 */
public class JNDIRegistry {

    /**
     * Name of the JNDI endpoints context
     */
    public static final String ENDPOINTS_REF = "endpoints";

    /**
     * Name of the JNDI interfaces context
     */
    public static final String INTERFACES_REF = "interfaces";

    /**
     * Name of the JNDI new endpoints context
     */
    public static final String NEW_ENDPOINTS_REF = "new-endpoints";

    /**
     * Name of the JNDI service context
     */
    public static final String SERVICES_REF = "services";

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Endpoint context
     */
    protected Context endpointsContext;

    /**
     * Interfaces context
     */
    protected Context interfacesContext;

    /**
     * New endpoint context
     */
    protected Context newEndpointContext;

    /**
     * Root context
     */
    protected InitialContext rootContext;

    /**
     * Service context
     */
    protected Context servicesContext;

    /**
     * JNDI service fractal component
     */
    protected JNDIService jndiService;

    /**
     * Configuration service fractal component
     */
    protected ConfigurationService configurationService;

    public JNDIRegistry(ConfigurationService configurationService, JNDIService jndiService,
            LoggingUtil log) {
        this.log = log;
        this.configurationService = configurationService;
        this.jndiService = jndiService;
    }

    /**
     * @see Registry#cleanNewEndpoints()
     */
    public void cleanNewEndpoints() throws RegistryException {
        NamingEnumeration<NameClassPair> newEndpoints;

        try {
            // get the new endpoints
            newEndpoints = this.rootContext.list(NEW_ENDPOINTS_REF);
        } catch (NamingException e) {
            throw new RegistryException("Problem while retrieving new endpoints context", e);
        }

        try {
            // unbind all the bound objects
            while (newEndpoints.hasMore()) {
                NameClassPair nameClassPair = newEndpoints.next();
                try {
                    this.newEndpointContext.unbind(nameClassPair.getName());
                } catch (NamingException e) {
                    this.log.warning("Failed to unbind the new endpoint '"
                            + nameClassPair.getName() + "'");
                }
            }
        } catch (NamingException e) {
            throw new RegistryException("Problem while cleaning the new endpoints", e);
        }

    }

    /**
     * Unregister a connection
     * 
     * @param connection
     *            connection to unregistered; must be non null
     * @throws RegistryException
     *             if the connection was not previously registered
     */
    public void unregisterConnection(LinkedEndpoint linkedEndpoint) throws RegistryException {
        this.log.start();
        if (linkedEndpoint == null) {
            throw new IllegalArgumentException("The connection must be non null");
        }

        String toKey = getKey(linkedEndpoint.getToServiceName(), linkedEndpoint.getToEndpointName());

        String key = getKey(linkedEndpoint.getServiceName(), linkedEndpoint.getEndpointName());

        if (linkedEndpoint.getInterfaces() != null) {
            Name interf = NamingHelper.qNameToName(linkedEndpoint.getInterfaces()[0]);
            try {
                if (!JNDIUtil.isBound(this.rootContext, INTERFACES_REF, interf)) {
                    throw new RegistryException("The connection interface name "
                            + linkedEndpoint.getInterfaces()[0] + " is not registered");
                } else {

                    // check if the provider endpoint is still bound in the
                    // new endpoint context if so its interfaces have not
                    // been resolved
                    if (JNDIUtil.isBound(this.rootContext, NEW_ENDPOINTS_REF, toKey)) {
                        Context interfaceCtx = (Context) this.interfacesContext.lookup(interf);

                        interfaceCtx.unbind(key);
                    } else {
                        // we need to check that we do not remove it from
                        // one of its interfaces context
                        InternalEndpoint internalEndpoint = (InternalEndpoint) this.getInternalEndpoint(
                                linkedEndpoint.getToServiceName(), linkedEndpoint
                                        .getToEndpointName());
                        boolean isNotInInt = true;
                        for (QName i : internalEndpoint.getInterfaces()) {
                            if (linkedEndpoint.getInterfaces()[0].equals(i)) {
                                isNotInInt = false;
                            }
                        }
                        if (isNotInInt) {
                            Context interfaceCtx = (Context) this.interfacesContext.lookup(interf);
                            interfaceCtx.unbind(key);
                            if (!this.interfacesContext.list(interf).hasMoreElements()) {
                                this.interfacesContext.destroySubcontext(interf);
                            }
                        }
                    }
                }
            } catch (NamingException e) {
                throw new RegistryException(
                        "problem while unregistering connection with interfacename "
                                + linkedEndpoint.getInterfaces()[0] + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        } else {
            try {
                if (!JNDIUtil.isBound(this.rootContext, ENDPOINTS_REF, key)) {
                    throw new RegistryException("The provider endpoint '" + key
                            + "' is not registered in the endpoints");
                }

                // If the prov service name is different from the cons service
                // name we removed it
                if (!linkedEndpoint.getToServiceName().equals(linkedEndpoint.getServiceName())) {
                    this.retrieveServiceContext(
                            NamingHelper.qNameToName(linkedEndpoint.getServiceName())).unbind(key);
                    // If the service context is empty, we need to remove it
                    Name serviceName = NamingHelper.qNameToName(linkedEndpoint.getServiceName());
                    if (!this.servicesContext.list(serviceName).hasMore()) {
                        this.servicesContext.destroySubcontext(serviceName);
                    }
                }

                // If the prov endpoint name is different from the cons endpoint
                // name we removed it
                // FIXME : Test the service name too ???
                if (!toKey.equals(key)) {
                    this.endpointsContext.unbind(key);
                }

            } catch (NamingException e) {
                throw new RegistryException(
                        "problem while registering connection with service name "
                                + linkedEndpoint.getServiceName() + " and endpoint name "
                                + linkedEndpoint.getEndpointName() + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        }
        this.log.end();
    }

    /**
     * Deregister the given external endpoint.
     * 
     * @param externalEndpoint
     *            the external endpoint to be deregistered; must be non-null
     * @throws RegistryException
     *             if the endpoint was not previously registered
     */
    public void unregisterExternalEndpoint(ServiceEndpoint externalEndpoint)
            throws RegistryException {
        this.log.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException("The external endpoint must be non null");
        }
        this.unregisterInternalEndpoint((AbstractEndpoint) externalEndpoint);
        this.log.end();
    }

    /**
     * @see Registry#unregisterInternalEndpoint(AbstractEndpoint)
     *      <ol>
     *      <li> Unbind the endpoint from the endpoint context</li>
     *      <li> If the endpoint is still in the new-endpoints context, unbind
     *      the endpoint from the new-endpoints context</li>
     *      <li> Unbind the endpoint from the service context with contains it</li>
     *      <li> If the service subcontext is empty, remove it from the services
     *      context</li>
     *      <li> Unbind the endpoint from the interface context with contains it</li>
     *      <li> If the interface subcontext is empty, remove it from the
     *      interfaces context</li>
     *      </ol>
     */
    public void unregisterInternalEndpoint(AbstractEndpoint internalEndpoint)
            throws RegistryException {
        this.log.start();
        if (internalEndpoint == null) {
            throw new IllegalArgumentException("The endpoint must be non null");
        }

        String key = this.getKey(internalEndpoint);

        try {
            // Check if the endpoint is already registered before unregistering
            // it
            if (!JNDIUtil.isBound(this.rootContext, ENDPOINTS_REF, key)) {
                throw new RegistryException("The endpoint " + internalEndpoint.getEndpointName()
                        + " must be previously registered to be unregistered");
            }

            if (JNDIUtil.isBound(this.rootContext, NEW_ENDPOINTS_REF, key)) {
                // 1.1 Unbind from new-endpoints context
                this.newEndpointContext.unbind(key);
            } else if (internalEndpoint.getInterfaces() != null) {
                // 1.2 or Unbind from interfaces context
                for (QName interfaceName : internalEndpoint.getInterfaces()) {
                    Name itfName = NamingHelper.qNameToName(interfaceName);
                    if (JNDIUtil.isBound(this.rootContext, INTERFACES_REF, itfName)) {
                        Context interfaceCtx = (Context) this.interfacesContext.lookup(itfName);

                        interfaceCtx.unbind(key);

                        // delete subcontext if no more endpoint is found
                        if (!this.interfacesContext.list(itfName).hasMoreElements()) {
                            this.interfacesContext.destroySubcontext(itfName);
                        }
                    }
                }
            }

            //
            // 2.Unbind from services context
            if (!StringHelper.isNullOrEmpty(NamingHelper.qNameToString(internalEndpoint
                    .getServiceName()))) {
                Name serviceName = NamingHelper.qNameToName(internalEndpoint.getServiceName());

                if (JNDIUtil.isBound(this.rootContext, SERVICES_REF, serviceName)) {
                    Context serviceCtx = (Context) this.servicesContext.lookup(serviceName);

                    serviceCtx.unbind(key);

                    // destroy service subcontext if no more endpoints serve
                    // this services
                    if (!this.servicesContext.list(serviceName).hasMore()) {
                        this.servicesContext.destroySubcontext(serviceName);
                    }
                } else {
                    throw new RegistryException("Could not deregister internal endpoint "
                            + internalEndpoint.getEndpointName() + " its service "
                            + internalEndpoint.getServiceName() + " is not registered");
                }
            }

            // 3.Remove the element from endpoints map
            this.endpointsContext.unbind(key);

        } catch (NamingException e) {
            throw new RegistryException(e);
        }
        this.log.end();
    }

    /**
     * Get the external service endpoint for the named activated endpoint, if
     * any
     * 
     * @param service
     *            qualified name of the endpoint's service; must be non null
     * @param name
     *            name of the endpoint; must be non null
     * @return the named external endpoint, or null if the named endpoint is not
     *         activated
     */
    public AbstractEndpoint getExternalEndpoint(QName serviceName, String endpointName)
            throws RegistryException {
        this.log.call();
        AbstractEndpoint result = this.retrieveEndpoint(serviceName, endpointName,
                ExternalEndpoint.class.getName());
        if (result != null) {
            if (!this.isVisibleEndpoint(result)) {
                result = null;
            }
        }
        return result;
    }

    /**
     * Queries for external endpoints that implements the given interface
     * 
     * @param interfaceName
     *            qualified name of the interface that contains the endpoints
     * @return Array of external endpoints
     * 
     * Array of external endpoints that implements the qualified interface, if
     * the interface name is non null
     * 
     * Array of all the external endpoints, if the interface name is null
     * 
     * @throws RegistryException
     *             if a problem occurs
     */
    public AbstractEndpoint[] getExternalEndpointsForInterface(QName interfaceName)
            throws RegistryException {
        this.log.call();
        return this.removeNonVisibleEndpoints(this.retrieveEndpointsForInterface(interfaceName,
                ExternalEndpoint.class.getName()));
    }

    /**
     * Queries for external endpoints that are part of the given service
     * 
     * @param serviceName
     *            qualified name of the service that contains the endpoints;
     *            must be non null
     * 
     * @return array of available external endpoints for the specified service
     *         name; must be non null; may be empty
     * 
     */
    public AbstractEndpoint[] getExternalEndpointsForService(QName serviceName) {
        this.log.call();
        return this.removeNonVisibleEndpoints(this.retrieveEndpointsForService(serviceName,
                ExternalEndpoint.class.getName()));
    }

    /**
     * Get the internal service endpoint for the named activated endpoint, if
     * any. if the endpoint is a link, resolve it and return the real endpoint
     * 
     * @param service
     *            qualified name of the endpoint's service; must be non null
     * @param name
     *            name of the endpoint; must be non null
     * @return the named internal endpoint, or null if the named endpoint is not
     *         activated
     */
    public AbstractEndpoint getInternalEndpoint(QName serviceName, String endpointName)
            throws RegistryException {
        this.log.call();
        AbstractEndpoint result = this.retrieveEndpoint(serviceName, endpointName,
                InternalEndpoint.class.getName());
        if (result != null) {
            if (!this.isVisibleEndpoint(result)) {
                result = null;
            }
        }
        return result;
    }

    /**
     * Get the internal service endpoint for the named activated endpoint, if
     * any
     * 
     * @param service
     *            qualified name of the endpoint's service; must be non null
     * @param name
     *            name of the endpoint; must be non null
     * @param resolveLink
     *            if the Endpoint is a link, resolve it (true) or return the
     *            link (false)
     * @return the named internal endpoint, or null if the named endpoint is not
     *         activated
     */
    public AbstractEndpoint getInternalEndpoint(QName service, String name, boolean resolveLink)
            throws RegistryException {
        this.log.call();
        AbstractEndpoint result = this.retrieveEndpoint(service, name, InternalEndpoint.class.getName(),
                resolveLink);
        if (result != null) {
            if (!this.isVisibleEndpoint(result)) {
                result = null;
            }
        }
        return result;
    }

    /**
     * Queries for internal endpoints that implements the given interface
     * 
     * @param interfaceName
     *            qualified name of the interface that contains the endpoints
     * @return Array of internal endpoints
     * 
     * Array of internal endpoints that implements the qualified interface, if
     * the interface name is non null
     * 
     * Array of all the external endpoints, if the interface name is null
     * 
     * @throws RegistryException
     *             if a problem occurs
     */
    public AbstractEndpoint[] getInternalEndpointsForInterface(QName interfaceName)
            throws RegistryException {
        this.log.call();
        return this.removeNonVisibleEndpoints(this.retrieveEndpointsForInterface(interfaceName,
                InternalEndpoint.class.getName()));
    }

    /**
     * Queries for internal endpoints that are part of the given service
     * 
     * @param serviceName
     *            qualified name of the service that contains the endpoints;
     *            must be non null
     * 
     * @return array of available internal endpoints for the specified service
     *         name; must be non null; may be empty
     * 
     */
    public AbstractEndpoint[] getInternalEndpointsForService(QName serviceName) {
        this.log.call();
        return this.removeNonVisibleEndpoints(this.retrieveEndpointsForService(serviceName,
                InternalEndpoint.class.getName()));
    }

    /**
     * Register a connection
     * 
     * @param connection
     *            connection to registered; must be non null
     * @throws RegistryException
     *             if the connection has already been registered or if a problem
     *             occurs
     */
    public void registerConnection(LinkedEndpoint linkedEndpoint) throws RegistryException {
        this.log.start();
        if (linkedEndpoint == null) {
            throw new IllegalArgumentException("The connection must be non null");
        }

        // targeted endpoint
        String toKey = getKey(linkedEndpoint.getToServiceName(), linkedEndpoint.getToEndpointName());

        // 'virtual' endpoint
        String key = getKey(linkedEndpoint.getServiceName(), linkedEndpoint.getEndpointName());

        LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/" + toKey);

        if (linkedEndpoint.getInterfaces() != null) {
            try {
                Name itfName = NamingHelper.qNameToName(linkedEndpoint.getInterfaces()[0]);
                Context itfContext = this.retrieveInterfaceContext(itfName);

                // bind or rebind link in interfaces
                if (!JNDIUtil.isBound(this.rootContext, INTERFACES_REF, itfName)) {
                    itfContext.bind(key, linkRef);
                } else {
                    itfContext.rebind(key, linkRef);
                }
            } catch (NamingException e) {
                throw new RegistryException(
                        "Problem while registering connection with interfacename "
                                + linkedEndpoint.getInterfaces()[0] + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        } else {
            try {
                Name serviceName = NamingHelper.qNameToName(linkedEndpoint.getServiceName());
                Context servContext = this.retrieveServiceContext(serviceName);

                // bind or rebind service
                if (!JNDIUtil.isBound(this.rootContext, SERVICES_REF, key)) {
                    servContext.bind(key, linkRef);
                } else {
                    servContext.rebind(key, linkRef);
                }

                // bind or rebind endpoint
                if (!JNDIUtil.isBound(this.rootContext, ENDPOINTS_REF, key)) {
                    this.endpointsContext.bind(key, linkRef);
                } else {
                    this.endpointsContext.rebind(key, linkRef);
                }

            } catch (NamingException e) {
                throw new RegistryException(
                        "Problem while registering connection with service name  "
                                + linkedEndpoint.getServiceName() + " and endpoint name "
                                + linkedEndpoint.getEndpointName() + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        }
        this.log.end();
    }

    /**
     * Register the given external endpoint
     * 
     * @param externalEndpoint
     *            the external endpoint to be registered; must be non null
     * @throws RegistryException
     *             if an external endpoint with the same name is already
     *             registered
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint) throws RegistryException {
        this.log.start();
        this.registerEndpoint(externalEndpoint);
        this.log.end();
    }

    /**
     * Register the given internal endpoint
     * 
     * @param internalEndpoint
     *            the internal endpoint to be registered; must be non null
     * @throws RegistryException
     *             if an internal endpoint with the same name is already
     *             registered
     */
    public void registerInternalEndpoint(AbstractEndpoint endpoint) throws RegistryException {
        this.log.start();
        this.registerEndpoint(endpoint);
        this.log.end();
    }

    /**
     * Retreive the list of all the new endpoints
     * 
     * @return list of new endpoints
     * @throws RegistryException
     *             if a problem occurs
     */
    public List<AbstractEndpoint> retrieveNewEndpoints() throws RegistryException {
        this.log.start();

        List<AbstractEndpoint> endpoints = new ArrayList<AbstractEndpoint>();
        try {
            NamingEnumeration<Binding> bindings = this.rootContext.listBindings(NEW_ENDPOINTS_REF);

            while (bindings.hasMoreElements()) {
                Binding binding = bindings.next();
                Object object = binding.getObject();
                if (object instanceof LinkRef) {
                    AbstractEndpoint abstractEndpoint = (AbstractEndpoint) this.rootContext
                            .lookup(((LinkRef) object).getLinkName());
                    endpoints.add(abstractEndpoint);
                } else {
                    throw new RegistryException("The new endpoint entry named '"
                            + binding.getName() + "' must be a " + LinkRef.class + " instance");
                }
            }
        } catch (Exception e) {
            throw new RegistryException("Problem while retrieving all the new endpoints", e);
        }

        this.log.end();
        return endpoints;
    }

    /**
     * Update the endpoint, remove it from the new-endpoint context and update
     * the interfaces context
     * 
     * @param endpoint
     *            endpoint to validate; must be non null
     * @throws RegistryException
     *             if a problem occurs
     */
    public void validateEndpoint(AbstractEndpoint endpoint) throws RegistryException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("The endpoint must be non null");
        }
        String key = this.getKey(endpoint);

        try {
            // Add the interfaces to the interfaces context
            for (QName name : endpoint.getInterfaces()) {
                Name itfName = NamingHelper.qNameToName(name);

                Context interfaceCtx = this.retrieveInterfaceContext(itfName);
                LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/" + key);

                // bind or rebind endpoint in interfaces
                if (JNDIUtil.isBound(this.interfacesContext, itfName, key)) {
                    interfaceCtx.rebind(key, linkRef);
                } else {
                    interfaceCtx.bind(key, linkRef);
                }
            }

            // Update the endpoint in the endpoint context
            this.endpointsContext.rebind(key, endpoint);

        } catch (Exception e) {
            throw new RegistryException(e);
        }
        this.log.end();
    }

    /**
     * Bind the endpoint in the endpointcontext
     * 
     * @param endpoint
     *            endpoint to bind
     * @throws NamingException
     */
    @SuppressWarnings("unchecked")
    protected void bindInEndpointContext(AbstractEndpoint endpoint) throws NamingException {
        this.log.call();
        String key = this.getKey(endpoint);
        if (JNDIUtil.isBound(this.rootContext, ENDPOINTS_REF, key)) {
            throw new NamingException("Endpoint " + endpoint.getEndpointName()
                    + " is already bound");
        } else {
            this.endpointsContext.bind(key, endpoint);
        }
    }

    /**
     * Bind the endpoint in the newendpointcontext
     * 
     * @param endpoint
     *            endpoint to bind
     * @throws NamingException
     */
    protected void bindInNewEndpointContext(AbstractEndpoint endpoint) throws NamingException {
        this.log.call();
        String key = this.getKey(endpoint);
        LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/" + key);
        if (!JNDIUtil.isBound(this.rootContext, NEW_ENDPOINTS_REF, key)) {
            this.newEndpointContext.bind(key, linkRef);
        }
    }

    /**
     * Bind the endpoint in the servicecontext. If no serviceName is specified,
     * do nothing.
     * 
     * @param endpoint
     *            endpoint to bind
     * @throws NamingException
     */
    protected void bindInServiceContext(AbstractEndpoint endpoint) throws NamingException {
        this.log.call();
        String key = this.getKey(endpoint);

        if (!StringHelper.isNullOrEmpty(NamingHelper.qNameToString(endpoint.getServiceName()))) {
            Name serviceName = NamingHelper.qNameToName(endpoint.getServiceName());

            LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/" + key);
            Context servContext = this.retrieveServiceContext(serviceName);
            if (!JNDIUtil.isBound(this.servicesContext, serviceName, key)) {
                servContext.bind(key, linkRef);
            }
        }
    }

    /**
     * Create and / or reference contexts "containers", "endpoints", "services"
     * and "interfaces" in the global directory.
     * 
     * @throws JBIException
     *             problems while creating or lookup contexts
     */
    public void initContexts() throws RegistryException {
        this.log.start();
        try {
            this.rootContext = this.jndiService.getInitialContext();

            if (JNDIUtil.isBound(this.rootContext, "/", ENDPOINTS_REF)) {
                this.endpointsContext = (Context) this.rootContext.lookup(ENDPOINTS_REF);
            } else {
                this.endpointsContext = this.rootContext.createSubcontext(ENDPOINTS_REF);
            }
            if (JNDIUtil.isBound(this.rootContext, "/", SERVICES_REF)) {
                this.servicesContext = (Context) this.rootContext.lookup(SERVICES_REF);
            } else {
                this.servicesContext = this.rootContext.createSubcontext(SERVICES_REF);
            }
            if (JNDIUtil.isBound(this.rootContext, "/", NEW_ENDPOINTS_REF)) {
                this.newEndpointContext = (Context) this.rootContext.lookup(NEW_ENDPOINTS_REF);
            } else {
                this.newEndpointContext = this.rootContext.createSubcontext(NEW_ENDPOINTS_REF);
            }
            if (JNDIUtil.isBound(this.rootContext, "/", INTERFACES_REF)) {
                this.interfacesContext = (Context) this.rootContext.lookup(INTERFACES_REF);
            } else {
                this.interfacesContext = this.rootContext.createSubcontext(INTERFACES_REF);
            }

            if (this.endpointsContext == null || this.servicesContext == null || this.interfacesContext == null) {
                throw new NotContextException();
            }
        } catch (Exception e) {
            throw new RegistryException(e);
        }
        this.log.end();
    }

    /**
     * @see Registry#registerExternalEndpoint(ServiceEndpoint)
     */
    protected void registerEndpoint(ServiceEndpoint endpoint) throws RegistryException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must not be null");
        }

        String key = this.getKey(endpoint);

        try {
            if (JNDIUtil.isBound(this.rootContext, ENDPOINTS_REF, key)) {
                // Retrieve a potential matching endpoint
                AbstractEndpoint registeredEndpoint = this.retrieveEndpoint(endpoint);

                // registeredEndpoint == null ==> the bound object isn't a real
                // endpoint (link or anything else)
                if (registeredEndpoint == null) {
                    throw new RegistryException(
                            "You are trying to register an endpoint with a name that "
                                    + "is already bound to an other object "
                                    + "(link or anything else) in the JNDI Registry.");
                }

                // An endpoint can't be registered with the same endpoint name
                // as a
                // previously registered endpoint that have different container
                // name
                // or component name
                if (!(((AbstractEndpoint) endpoint).getContainerName().equals(
                        registeredEndpoint.getContainerName()) && ((AbstractEndpoint) endpoint)
                        .getComponentName().equals(registeredEndpoint.getComponentName()))) {
                    throw new RegistryException("The endpoint " + endpoint.getEndpointName()
                            + " is already registered for an other component/container");
                }

            } else {
                this.bindInEndpointContext((AbstractEndpoint) endpoint);

                this.bindInNewEndpointContext((AbstractEndpoint) endpoint);

                this.bindInServiceContext((AbstractEndpoint) endpoint);
            }
        } catch (NamingException e) {
            throw new RegistryException("Problem while registering endpoint "
                    + endpoint.getEndpointName(), e);
        }
        this.log.end();
    }

    /**
     * Get the internal/external service endpoint for the named activated
     * endpoint, if any. If the Endpoint is a link, resolve it and return the
     * real endpoint
     * 
     * @param serviceName
     *            the {@link QName} of the service
     * @param name
     *            the endpoint name
     * @param endpointClassName
     *            class name of the endpoint
     * @return the matching endpoint
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint retrieveEndpoint(QName serviceName, String endpointName,
            String endpointClassName) throws RegistryException {
        this.log.call();
        AbstractEndpoint endpoint = this.retrieveEndpoint(serviceName, endpointName, endpointClassName,
                true);
        return endpoint;
    }

    /**
     * Get the internal/external service endpoint for the named activated
     * endpoint, if any
     * 
     * @param service
     *            qualified service name
     * @param endpointName
     *            endpoint name
     * @param endpointClassName
     *            class name of the endpoint
     * @param resolveLink
     *            if the Endpoint is a link, resolve it (true) or return the
     *            link (false)
     * @return
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint retrieveEndpoint(QName serviceName, String endpointName,
            String endpointClassName, boolean resolveLink) throws RegistryException {
        this.log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("The service name must be non null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("The endpoint name must be non null");
        }
        AbstractEndpoint result = null;
        String key = getKey(serviceName, endpointName);

        try {

            Object outObj = this.endpointsContext.lookup(key);

            // real endpoint
            if (outObj instanceof AbstractEndpoint) {
                AbstractEndpoint endpoint = (AbstractEndpoint) outObj;
                if (endpoint.getClass().getName().equals(endpointClassName)) {
                    result = endpoint;
                }
            }
            // link endpoint
            else if (outObj instanceof LinkRef) {
                Object obj = this.rootContext.lookup(((LinkRef) outObj).getLinkName());
                if (obj instanceof AbstractEndpoint) {
                    AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                    if (endpoint.getClass().getName().equals(endpointClassName)) {
                        // return the real endpoint
                        if (resolveLink) {
                            result = endpoint;
                        }
                        // return a link
                        else {
                            result = new LinkedEndpoint(serviceName, endpointName, endpoint
                                    .getServiceName(), endpoint.getEndpointName(), null);
                        }
                    }
                }
            }
        } catch (NamingException e1) {
            // Do nothing
        }
        this.log.end();
        return result;
    }

    /**
     * Retrieve an endpoint in the JNDI registry without making distinction
     * between Internal and External one.
     * 
     * @param endpoint
     *            endpoint description
     * @return an {@link AbstractEndpoint} if the matching endpoint is a real
     *         endpoint, null otherwise (for exemple, the matching endpoint is a
     *         link)
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint retrieveEndpoint(ServiceEndpoint endpoint) throws RegistryException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("The endpoint must be non null");
        }

        AbstractEndpoint result = null;
        String key = this.getKey(endpoint);
        try {
            Object outObj = this.endpointsContext.lookup(key);

            // real endpoint
            if (outObj instanceof AbstractEndpoint) {
                result = (AbstractEndpoint) outObj;
            }
        } catch (NamingException e1) {
            throw new RegistryException("Error while looking up an Endpoint in the JNDI Registry",
                    e1);
        }
        this.log.end();
        return result;
    }

    /**
     * Retreive the list of endpoint that implements the interface name
     * 
     * @param interfaceName
     *            the {@link QName} of the interface
     * @param endpointType
     *            the type (class name) of the endpoint (InternalEndpoint or
     *            ExternalEndpoint)
     * @return an Array of matching endpoints
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint[] retrieveEndpointsForInterface(QName interfaceName,
            String endpointType) throws RegistryException {
        this.log.start();
        List<AbstractEndpoint> result = new ArrayList<AbstractEndpoint>();
        try {
            if (interfaceName == null) {
                // We must return all the endpoints of the specified type
                NamingEnumeration<Binding> bindings = this.rootContext.listBindings(ENDPOINTS_REF);

                while (bindings.hasMoreElements()) {
                    Binding binding = bindings.next();
                    Object object = binding.getObject();
                    // real endpoint case
                    if (object instanceof AbstractEndpoint) {
                        AbstractEndpoint endpoint = (AbstractEndpoint) object;
                        if (endpoint.getClass().getName().equals(endpointType)) {
                            result.add(endpoint);
                        }
                    }
                    // connection endpoint case
                    else if (object instanceof LinkRef) {
                        LinkRef linkRef = (LinkRef) object;
                        Object obj = this.rootContext.lookup(linkRef.getLinkName());
                        // connection linked to a real EP
                        if (obj instanceof AbstractEndpoint) {
                            AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                            if (endpoint.getClass().getName().equals(endpointType)) {
                                LinkedEndpoint epLink = new LinkedEndpoint(
                                        getServiceNameFromKey(binding.getName()),
                                        getEndpointNameFromKey(binding.getName()), endpoint
                                                .getServiceName(), endpoint.getEndpointName(), null);
                                result.add(epLink);
                            }
                        }
                    }
                }
            } else {
                // We must return only the endpoints for the specified
                // interface
                Name itfName = NamingHelper.qNameToName(interfaceName);
                if (JNDIUtil.isBound(this.rootContext, INTERFACES_REF, itfName)) {
                    NamingEnumeration<Binding> bindings = this.interfacesContext.listBindings(itfName);

                    while (bindings.hasMoreElements()) {
                        Binding binding = bindings.next();
                        Object object = binding.getObject();
                        if (object instanceof LinkRef) {
                            LinkRef linkRef = (LinkRef) object;
                            Object obj = this.rootContext.lookup(linkRef.getLinkName());
                            if (obj instanceof AbstractEndpoint) {
                                AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                                if (endpoint.getClass().getName().equals(endpointType)) {
                                    result.add(endpoint);
                                }
                            }
                        }
                    }
                }
            }
        } catch (NamingException e) {
            throw new RegistryException(
                    "Problem while getting endpoints for the specified interface " + interfaceName,
                    e);
        }
        this.log.end();
        return result.toArray(new AbstractEndpoint[0]);
    }

    /**
     * @see Registry#getExternalEndpointsForService(QName)
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint[] retrieveEndpointsForService(QName serviceName, String endpointType) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("The service name must be non null");
        }

        AbstractEndpoint[] result = new AbstractEndpoint[0];
        Name srvName = NamingHelper.qNameToName(serviceName);
        try {
            if (JNDIUtil.isBound(this.rootContext, SERVICES_REF, srvName)) {

                NamingEnumeration<Binding> bindings = this.servicesContext.listBindings(srvName);
                List<AbstractEndpoint> endpoints = new ArrayList<AbstractEndpoint>();
                while (bindings.hasMoreElements()) {
                    Binding binding = bindings.next();
                    Object object = binding.getObject();
                    if (object instanceof LinkRef) {
                        LinkRef linkRef = (LinkRef) object;
                        Object obj = this.rootContext.lookup(linkRef.getLinkName());
                        if (obj instanceof AbstractEndpoint) {
                            AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                            if (endpoint.getClass().getName().equals(endpointType)) {
                                endpoints.add(endpoint);
                            }
                        }
                    }
                }
                result = endpoints.toArray(new AbstractEndpoint[0]);
            }
        } catch (NamingException e) {
            this.log.error(e.getMessage(), e);
        }

        this.log.end();
        return result;
    }

    /**
     * Return the interfaceContext associated with this interfaceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    protected Context retrieveInterfaceContext(Name interfaceName) throws NamingException {
        this.log.call();
        Context interfContext = null;
        if (JNDIUtil.isBound(this.rootContext, INTERFACES_REF, interfaceName)) {
            interfContext = (Context) this.interfacesContext.lookup(interfaceName);
        } else {
            interfContext = this.interfacesContext.createSubcontext(interfaceName);
        }
        return interfContext;
    }

    /**
     * Return the serviceContext associated with this serviceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    protected Context retrieveServiceContext(Name serviceName) throws NamingException {
        this.log.call();
        Context serviceContext = null;
        if (JNDIUtil.isBound(this.rootContext, SERVICES_REF, serviceName)) {
            serviceContext = (Context) this.servicesContext.lookup(serviceName);
        } else {
            serviceContext = this.servicesContext.createSubcontext(serviceName);
        }
        return serviceContext;
    }

    /**
     * Get a unique key for the service endpoint
     * 
     * @param ep
     * @return
     */
    protected String getKey(ServiceEndpoint ep) {
        return getKey(ep.getServiceName(), ep.getEndpointName());
    }

    /**
     * Create an unique key for the serviceName / endpointName
     * 
     * @return
     */
    public static String getKey(QName serviceName, String endpointName) {
        String result = serviceName + "@" + endpointName;
        return result.replaceAll("/", "&47;");
    }

    /**
     * Retrieve the service name from a jndi key
     * 
     * @return
     */
    public static QName getServiceNameFromKey(String key) {
        String[] result = key.split("@");
        return QName.valueOf(result[0].replaceAll("&47;", "/"));
    }

    /**
     * Retrieve the endpoint name from a jndi key
     * 
     * @return
     */
    public static String getEndpointNameFromKey(String key) {
        String[] result = key.split("@");
        return result[1].replaceAll("&47;", "/");
    }

    /**
     * Remove from the collection the endpoints that must not be shown, ie
     * private endpoints from other sub-domains
     * 
     * @return a collection of visible endpoints
     */
    private AbstractEndpoint[] removeNonVisibleEndpoints(AbstractEndpoint[] endpoints) {
        ArrayList<AbstractEndpoint> result = new ArrayList<AbstractEndpoint>();
        for (AbstractEndpoint abstractEndpoint : endpoints) {
            if (this.isVisibleEndpoint(abstractEndpoint)) {
                result.add(abstractEndpoint);
            }
        }
        return result.toArray(new AbstractEndpoint[0]);
    }

    /**
     * return true if the endpoint is Visible <br>
     * (same sub domain as the local container or public endpoint) <br>
     * public services does not contains "private" in the namespace
     * 
     * @param abstractEndpoint
     * @return
     */
    private boolean isVisibleEndpoint(AbstractEndpoint abstractEndpoint) {
        String subDomain = this.configurationService.getContainerConfiguration().getSubdomain();
        return subDomain.equals(abstractEndpoint.getSubDomainName())
                || abstractEndpoint.getServiceName().getNamespaceURI().indexOf("private") <= 0;
    }
}
