/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.routing;

import java.util.Iterator;
import java.util.LinkedList;

import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.jbi.messaging.routing.module.RouterModule;

/**
 * The global router module contains all the router modules. It is in charge of
 * calling all the modules in the good order.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.2
 * 
 */
public class GlobalRouterModule extends RouterModule {

    /**
     * List of all the modules to call. Modules are ordered in the order to be
     * called. TODO : Add JMX Bean to add modules at runtime.
     */
    private LinkedList<RouterModule> modules;

    /**
     * Creates a new instance of GlobalRouterModule
     * 
     * @param modules
     */
    public GlobalRouterModule(LinkedList<RouterModule> modules) {
        if (modules == null) {
            modules = new LinkedList<RouterModule>();
        }
        this.modules = modules;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#init()
     */
    @Override
    public void doInit() throws Exception {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#addComponent(org.ow2.petals.jbi.component.context.ComponentContextImpl)
     */
    @Override
    public void addComponent(ComponentContextImpl componentContext) throws RoutingException {
        for (Iterator<RouterModule> iterator = modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call addComponent for module " + routerModule.getName());
            try {
                routerModule.addComponent(componentContext);
            } catch (RoutingException e) {
                log.warning("Error on addComponent : " + e.getMessage());
                throw e;
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#receive(org.ow2.petals.jbi.component.context.ComponentContextImpl,
     *      long)
     */
    @Override
    public MessageExchangeImpl receive(ComponentContextImpl source, long timeoutMS)
            throws RoutingException {
        MessageExchangeImpl result = null;
        for (Iterator<RouterModule> iterator = modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call receive for module " + routerModule.getName());

            try {
                MessageExchangeImpl tmp = routerModule.receive(source, timeoutMS);
                if (tmp != null) {
                    // cache the last non null receive...
                    result = tmp;
                }
            } catch (RoutingException e) {
                log.warning("Error on receive: " + e.getMessage());
                throw e;
            }
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#removeComponent(org.ow2.petals.jbi.component.context.ComponentContextImpl)
     */
    @Override
    public void removeComponent(ComponentContextImpl componentContext) throws RoutingException {
        for (Iterator<RouterModule> iterator = modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call remove component for module " + routerModule.getName());

            try {
                routerModule.removeComponent(componentContext);
            } catch (RoutingException e) {
                log.warning("Error on remove component : " + e.getMessage());
                throw e;
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#send(org.ow2.petals.jbi.component.context.ComponentContextImpl,
     *      org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    @Override
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange, long timeOut)
            throws RoutingException {
        for (Iterator<RouterModule> iterator = modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call send for module " + routerModule.getName());

            try {
                routerModule.send(source, exchange, timeOut);
            } catch (RoutingException e) {
                log.warning("Error on send : " + e.getMessage());
                throw e;
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#sendSync(org.ow2.petals.jbi.component.context.ComponentContextImpl,
     *      org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    @Override
    public MessageExchangeImpl sendSync(ComponentContextImpl source, MessageExchangeImpl exchange,
            long timeOut) throws RoutingException {
        MessageExchangeImpl result = exchange;
        for (Iterator<RouterModule> iterator = modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call sendSync for module " + routerModule.getName());

            try {
                MessageExchangeImpl tmp = routerModule.sendSync(source, result, timeOut);
                if (tmp != null) {
                    result = tmp;
                }
            } catch (RoutingException e) {
                log.warning("Error on sendSync : " + e.getMessage());
                throw e;
            }
        }
        return result;
    }
}
