/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.routing.module;

import java.net.URISyntaxException;
import java.net.URL;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Set;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.security.auth.Subject;

import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.petals.jaas.GroupPrincipal;
import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.jbi.security.AuthorizationEntry;
import org.ow2.petals.jbi.security.AuthorizationException;
import org.ow2.petals.jbi.security.AuthorizationMap;
import org.ow2.petals.jbi.security.DefaultAuthorizationMap;
import org.ow2.petals.jbi.security.DefaultAuthorizationParser;

/**
 * A router module for authorizing routing. The accessible
 * endpoint/service/operation are defined as accessible in a configuration file
 * (authorization.cfg). This file must be available in the petals
 * classpath(generally under the conf directory). The incoming message must
 * contains a security subject where principals match with the defined ones.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.2
 * 
 */
public class AuthorizationModule extends RouterModule {

    public static final String CONFIGURATION = "authorization.cfg";

    private AuthorizationMap authorizationMap;

    /**
     * Creates a new instance of AuthorizationModule
     */
    public AuthorizationModule() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#init()
     */
    @Override
    protected void doInit() throws Exception {
        final URL authURL = this.getClass().getResource("/" + AuthorizationModule.CONFIGURATION);
        if (authURL == null) {
            throw new Exception("Can not load the authorization resource from classpath");
        }

        try {
            final DefaultAuthorizationParser parser = new DefaultAuthorizationParser(authURL
                    .toURI());
            this.authorizationMap = parser.parse();
        } catch (final URISyntaxException e) {
            throw new RoutingException("Authorization init fails", e);
        } catch (final AuthorizationException e) {
            throw new RoutingException("Authorization init fails", e);
        } finally {
            if (this.authorizationMap == null) {
                this.authorizationMap = new DefaultAuthorizationMap(
                        new ArrayList<AuthorizationEntry>(0));
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterFilter#send(org.ow2.petals.jbi.component.context.ComponentContextImpl,
     *      org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    @Override
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange, long timeOut)
            throws RoutingException {
        this.authorizeSend(source, exchange);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterFilter#sendSync(org.ow2.petals.jbi.component.context.ComponentContextImpl,
     *      org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    @Override
    public MessageExchangeImpl sendSync(ComponentContextImpl source, MessageExchangeImpl exchange,
            long timeOut) throws RoutingException {
        this.authorizeSend(source, exchange);
        return null;
    }

    /**
     * Authorize to send the message. This authorization is based on the message
     * exchange subject. If the subject principals match with the local
     * authorization map, the message can be send to the foreign
     * service/operation.
     * 
     * @param source
     * @param exchange
     * @param timeOut
     * @throws RoutingException
     *             if the subject is null or if the user is not authorized to
     *             send message to the JBI service (ie not found in
     *             authorization map).
     */
    protected void authorizeSend(ComponentContextImpl source, MessageExchangeImpl exchange)
            throws RoutingException {
        if ((exchange.getRole() == Role.CONSUMER) && (exchange.getStatus() != ExchangeStatus.DONE)) {
            final ServiceEndpoint se = exchange.getEndpoint();
            if (se != null) {
                final Set<Principal> acls = this.authorizationMap.getAccessControlList(se, exchange
                        .getOperation());

                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    for (Principal principal : acls) {
                        log.debug("Server defined ACL for operation : "
                                + exchange.getOperation().toString());
                        log.debug(principal.getName() + "(" + principal.getClass().getName() + ")");
                    }
                }

                // skip message if ACL contains ALL policy
                if (!acls.contains(GroupPrincipal.ALL)) {
                    final Subject subject = exchange.getMessage("in").getSecuritySubject();
                    if (subject == null) {
                        throw new SecurityException(
                                "User not authenticated (security subject is null)");
                    }

                    acls.retainAll(subject.getPrincipals());
                    log.debug("ACLS size after retain is : " + acls.size());

                    if (acls.size() == 0) {
                        throw new SecurityException("Endpoint '" + se.getEndpointName()
                                + "' / Operation '" + exchange.getOperation()
                                + "' is not authorized for this user");
                    }
                }
            }
        }
    }
}
