/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.routing.module;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.transport.TransportException;

/**
 * This is the router implementation which really send message to the
 * transporter layer.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.2
 * 
 */
public class RouterSenderModule extends RouterModule {

    /**
     * Creates a new instance of RouterSenderModule
     */
    public RouterSenderModule() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#init()
     */
    @Override
    protected void doInit() throws Exception {
    }

    /**
     * Sends the specified message. Depending on the Role of the sender, send
     * the message to the producer (request) or to the consumer (response)
     * 
     * @param source
     *            the source
     * @param exchange
     *            the exchange
     * @param timeOut
     *            timeout of the message (in ms). 0 for no timeout
     * @throws RoutingException
     *             impossible to send the message
     */
    public void send(final ComponentContextImpl source, final MessageExchangeImpl exchange,
            final long timeout) throws RoutingException {
        this.log.call();

        if (!checkBypassMessageExchange(exchange)) {
            if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
                this.sendToProvider(source, exchange, timeout, false);
            } else {
                this.sendToConsumer(source, exchange, timeout, false);
            }
        }
    }

    /**
     * Send the message synchronously into the NMR.
     * 
     * @param source
     *            the source
     * @param exchange
     *            the exchange
     * @param timeOut
     *            timeout of the message (in ms). 0 for no timeout
     * 
     * @return the response exchange if received, or <code>null</code> if the
     *         timeout is reached
     * 
     * @throws RoutingException
     *             impossible to process a synchronous send
     */
    public MessageExchangeImpl sendSync(final ComponentContextImpl source,
            final MessageExchangeImpl exchange, final long timeout) throws RoutingException {
        this.log.call();

        removeBypassMessageExchange(exchange);

        MessageExchangeImpl responseExchange = null;
        if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
            responseExchange = this.sendToProvider(source, exchange, timeout, true);
        } else {
            responseExchange = this.sendToConsumer(source, exchange, timeout, true);
        }

        return responseExchange;
    }

    /**
     * Receive a message from the Transporter and return it to the Delivery
     * Channel.
     * 
     * @param source
     *            The source component context
     * @param timeoutMS
     *            The timeout to receive the incoming message
     * @param block
     *            If <code>True</code>, the receive is blocking until a
     *            response
     * @throws RoutingException
     *             impossible to receive a message
     * @return the received message
     */
    public MessageExchangeImpl receive(final ComponentContextImpl source, final long timeoutMS)
            throws RoutingException {
        this.log.start();

        MessageExchangeImpl exchange = null;
        String componentName = source.getComponentName();

        try {
            long time = timeoutMS;
            exchange = this.transporter.receive(componentName, time);
        } catch (TransportException e) {
            throw new RoutingException(e);
        }

        this.log.end();

        return exchange;
    }

    /**
     * (non-Javadoc).
     * 
     * @see org.objectweb.petals.jbi.routing.Router#addComponent(java.lang.String)
     * @param componentName
     *            the name of the component
     * @throws RoutingException
     *             impossible to add the component
     */
    public void addComponent(final ComponentContextImpl componentContext) throws RoutingException {
        this.log.start();

        try {
            this.transporter.addDestination(componentContext.getComponentName());
        } catch (TransportException e) {
            throw new RoutingException(e);
        }

        this.log.end();
    }

    /**
     * (non-Javadoc).
     * 
     * @see org.objectweb.petals.jbi.routing.Router#removeComponent(java.lang.String)
     * @param componentName
     *            the name of the component
     * @throws RoutingException
     *             impossible to remove the component
     */
    public void removeComponent(final ComponentContextImpl componentContext)
            throws RoutingException {
        this.log.start();

        try {
            this.transporter.removeDestination(componentContext.getComponentName());
        } catch (TransportException e) {
            throw new RoutingException(e);
        }

        this.log.end();
    }

    /**
     * Sends the specified message, sent by the specified component. TODO send
     * directly to the DeliveryChannel of the component related to the recipient
     * endpoint.
     * 
     * @param source
     *            the source
     * @param exchange
     *            the exchange
     * @param timeOut
     *            timeout of the message (in ms). 0 for no timeout
     * @param sync
     *            true if the message is synchrone
     * @return the reponse exchange if a synchronious send is processed
     * @throws RoutingException
     *             impossible to send the message to the consumer
     */
    protected MessageExchangeImpl sendToConsumer(final ComponentContextImpl source,
            final MessageExchangeImpl exchange, final long timeout, final boolean sync)
            throws RoutingException {
        this.log.start();

        MessageExchangeImpl responseExchange = null;

        try {
            ConsumerEndpoint consumerEndpoint = exchange.getConsumerEndpoint();
            exchange.setRole(MessageExchange.Role.CONSUMER);

            if (sync) {
                responseExchange = this.transporter.sendSync(exchange, consumerEndpoint
                        .getComponentName(), consumerEndpoint.getContainerName(), timeout);
                // The timeout is reached
                if (responseExchange == null) {
                    exchange.setRole(MessageExchange.Role.PROVIDER);
                }
            } else {
                this.transporter.send(exchange, consumerEndpoint.getComponentName(),
                        consumerEndpoint.getContainerName(), timeout);
            }
        } catch (TransportException e) {
            // Restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.PROVIDER);
            throw new RoutingException(e);
        }

        this.log.end();
        return responseExchange;
    }

    /**
     * Sends the specified message to the provider of the service. TODO for the
     * moment, send directly to the DeliveryChannel of the component related to
     * the recipient endpoint.
     * 
     * @param source
     *            the source
     * @param exchange
     *            the exchange
     * @param timeOut
     *            timeout of the message (in ms). 0 for no timeout
     * @param sync
     *            true if the message is synchronous
     * @return the response exchange if a synchronous send is processed
     * @throws RoutingException
     *             impossible to send the message to the provider
     */
    protected MessageExchangeImpl sendToProvider(final ComponentContextImpl source,
            final MessageExchangeImpl exchange, final long timeOut, final boolean sync)
            throws RoutingException {
        this.log.start();

        MessageExchangeImpl responseExchange = null;

        // assume that address has been resolved in the AddressResolver module
        // this.resolveAddress(source, exchange);

        try {
            ServiceEndpoint providerEndpoint = exchange.getEndpoint();

            exchange.setRole(MessageExchange.Role.PROVIDER);

            if (sync) {
                responseExchange = this.transporter.sendSync(exchange,
                        ((AbstractEndpoint) providerEndpoint).getComponentName(),
                        ((AbstractEndpoint) providerEndpoint).getContainerName(), timeOut);
                // The timeout is reached
                if (responseExchange == null) {
                    exchange.setRole(MessageExchange.Role.CONSUMER);
                }
            } else {
                this.transporter.send(exchange, ((AbstractEndpoint) providerEndpoint)
                        .getComponentName(), ((AbstractEndpoint) providerEndpoint)
                        .getContainerName(), timeOut);
            }
        } catch (TransportException e) {
            // restore the Role if the message has not been sent
            exchange.setRole(MessageExchange.Role.CONSUMER);
            throw new RoutingException(e);
        }

        this.log.end();
        return responseExchange;
    }

    /**
     * Check if the message exchange to send is marked to be bypassed.
     * 
     * @param exchange
     *            The exchange to check
     */
    private boolean checkBypassMessageExchange(MessageExchangeImpl exchange) {
        this.log.call();

        boolean bypass = false;
        Object noAck = null;

        if (exchange.isTerminated()) {
            if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
                noAck = exchange.getProperty(PROPERTY_PROVIDER_NOACK);
            } else {
                noAck = exchange.getProperty(PROPERTY_CONSUMER_NOACK);
            }
            bypass = (noAck != null && noAck.toString().toLowerCase().equals("true"));
        }

        return bypass;
    }

    /**
     * If the message exchange to send is marked to be bypassed, remove the mark
     * and log a warning message.
     * 
     * @param exchange
     *            The exchange to check
     */
    private boolean removeBypassMessageExchange(MessageExchangeImpl exchange) {
        this.log.call();

        boolean bypass = false;

        if (exchange.isTerminated()) {
            if (MessageExchange.Role.CONSUMER.equals(exchange.getRole())) {
                if (exchange.getProperty(PROPERTY_CONSUMER_NOACK) != null) {
                    exchange.setProperty(PROPERTY_CONSUMER_NOACK, "false");
                    this.log.warning("Property '" + PROPERTY_CONSUMER_NOACK
                            + "' is not supported by a synchronous sending");
                }
            } else {
                if (exchange.getProperty(PROPERTY_PROVIDER_NOACK) != null) {
                    exchange.setProperty(PROPERTY_PROVIDER_NOACK, "false");
                    this.log.warning("Property '" + PROPERTY_PROVIDER_NOACK
                            + "' is not supported by a synchronous sending");
                }
            }
        }
        return bypass;
    }

}
