/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsAdminService.java 16:40:52 alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.kernel.admin;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean;
import org.ow2.petals.jbi.management.installation.InstallationServiceMBean;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.kernel.server.PetalsServer;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.util.LoggingUtil;

/**
 * This class is used for Petals global Administration
 * 
 * @author alouis - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.kernel.admin.PetalsAdminServiceInterface.class))
public class PetalsAdminServiceImpl implements PetalsAdminServiceInterface {

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * The Petals Server instance, which manages the Fractal composite
     */
    private PetalsServer petalsServer;

    /**
     * Deployment service fractal component
     */
    @Requires(name = "deployment", signature = org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean.class)
    protected DeploymentServiceMBean deployment;

    /**
     * Installation service fractal component
     */
    @Requires(name = "installation", signature = org.ow2.petals.jbi.management.installation.InstallationServiceMBean.class)
    protected InstallationServiceMBean installation;

    /**
     * Transporter service fractal component
     */
    @Requires(name = "transporter", signature = org.ow2.petals.transport.Transporter.class)
    protected Transporter transporter;

    /**
     * Topology service fractal component
     */
    @Requires(name = "topology", signature = org.ow2.petals.communication.topology.TopologyService.class, contingency = Contingency.OPTIONAL)
    protected TopologyService topologyService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * The domain configuration
     */
    private DomainConfiguration domainConfiguration;

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceMBean#retrieveContainersConfiguration()
     */
    public Set<Map<String, String>> retrieveContainersConfiguration() throws PetalsException {
        log.start();

        Set<Map<String, String>> containersSet = new HashSet<Map<String, String>>();

        Map<String, String> domainMap = this.getDomainMap();
        containersSet.add(domainMap);

        Set<ContainerConfiguration> containersConfiguration = retrieveContainersConfiguration(false);

        for (ContainerConfiguration container : containersConfiguration) {
            Map<String, String> containerMap = getContainerMap(container);
            containersSet.add(containerMap);
        }

        log.end();
        return containersSet;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceMBean#retrieveStartedContainersConfiguration()
     */
    public Set<Map<String, String>> retrieveStartedContainersConfiguration() throws PetalsException {
        log.start();

        Set<Map<String, String>> containersSet = new HashSet<Map<String, String>>();

        Map<String, String> domainMap = this.getDomainMap();
        containersSet.add(domainMap);

        Set<ContainerConfiguration> containersConfiguration = retrieveContainersConfiguration(true);

        for (ContainerConfiguration container : containersConfiguration) {
            Map<String, String> containerMap = getContainerMap(container);
            containersSet.add(containerMap);
        }

        log.end();
        return containersSet;
    }

    public void setPetalsServer(PetalsServer petalsServer) {
        this.petalsServer = petalsServer;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceMBean#shutdownContainer()
     */
    public void shutdownContainer() throws Exception {
        log.start();
        log.info("Shutting down the server");

        if (domainConfiguration.getMode().equals(DomainConfiguration.DomainMode.DYNAMIC)) {
            log.info("  1/5 Undeploy Service Assemblies...");
        } else {
            log.info("  1/4 Undeploy Service Assemblies...");
        }

        try {
            deployment.shutdown();
        } catch (Exception e) {
            log.error("Can not shut down the Deployment service.", e);
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {
            log.info("  2/5 Uninstall Shared Libraries and Components...");
        } else {
            log.info("  2/4 Uninstall Shared Libraries and Components...");
        }

        try {
            installation.shutdown();
        } catch (Exception e) {
            log.error("Can not shut down the Installation service", e);
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {
            log.info("  3/5 Shutting down the Transporter...");
        } else {
            log.info("  3/4 Shutting down the Transporter...");
        }

        try {
            transporter.shutdown();
        } catch (Exception e) {
            log.error("Can not shut down the Transporter service", e);
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {
            log.info("  4/5 Remove the PEtALS node from the domain...");
            topologyService.unregisterContainer(containerConfiguration.getName());
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {
            log.info("  5/5 Stopping the server...");
        } else {
            log.info("  4/4 Stopping the server...");
        }

        // launch the stop of the Petals server in a different thread
        petalsServer.getPetalsStopThread().start();

        log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceMBean#stopContainer()
     */
    public void stopContainer() throws Exception {
        log.start();
        log.info("Stopping the server...");
        // launch the stop of the Petals server in a different thread
        petalsServer.getPetalsStopThread().start();
        log.end();
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
        log = new LoggingUtil(logger);
        log.start();

        containerConfiguration = configurationService.getContainerConfiguration();
        domainConfiguration = configurationService.getDomainConfiguration();

        log.end();
    }

    /**
     * Retrieve the collection of all the servers configuration of this domain.
     * 
     * @param startedOnly
     *            return only started servers information
     * @return Set of server information ContainerInformation
     * @throws PetalsException
     */
    private Set<ContainerConfiguration> retrieveContainersConfiguration(boolean startedOnly)
            throws PetalsException {
        log.start();

        Set<ContainerConfiguration> containers = null;

        if (topologyService == null) {
            // it is the standalone distribution
            containers = new HashSet<ContainerConfiguration>();
            containers.add(containerConfiguration);
        } else {

            if (startedOnly) {
                containers = topologyService
                        .getAllContainersConfiguration(ContainerConfiguration.ContainerState.STARTED);
            } else {
                containers = topologyService.getAllContainersConfiguration(null);
            }
        }

        log.end();
        return containers;
    }

    /**
     * Get the domain configuration in key/value form.
     * 
     * @return
     */
    private Map<String, String> getDomainMap() {
        Map<String, String> domainMap = new Hashtable<String, String>();
        domainMap.put(CONF_DOMAIN_NAME, domainConfiguration.getDomain());
        if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {
            domainMap.put(CONF_DOMAIN_MODE, "dynamic");
        } else if (DomainConfiguration.DomainMode.STATIC.equals(domainConfiguration.getMode())) {
            domainMap.put(CONF_DOMAIN_MODE, "static");
        } else {
            domainMap.put(CONF_DOMAIN_MODE, "standalone");
        }

        if (domainConfiguration.getNetworkMuticastIp() != null) {
            domainMap.put(CONF_NETWORK_MULTICAST_IP, domainConfiguration.getNetworkMuticastIp());
            domainMap.put(CONF_NETWORK_MULTICAST_PORT, Integer.toString(domainConfiguration
                    .getNetworkMuticastPort()));
        }

        if (domainConfiguration.getJndiFactory() != null) {
            domainMap.put(CONF_JNDI_FACTORY, domainConfiguration.getJndiFactory());
        }
        if (domainConfiguration.getJndiHost() != null) {
            domainMap.put(CONF_JNDI_HOST, domainConfiguration.getJndiHost());
        }
        if (domainConfiguration.getJndiPort() != 0) {
            domainMap.put(CONF_JNDI_PORT, Integer.toString(domainConfiguration.getJndiPort()));
        }
        return domainMap;
    }

    /**
     * Get the container configuration in key/value form.
     * 
     * @param container
     * @return
     */
    private Map<String, String> getContainerMap(ContainerConfiguration container) {
        Map<String, String> containerMap = new Hashtable<String, String>();
        containerMap.put(CONF_CONTAINER_NAME, container.getName());
        containerMap.put(CONF_SUBDOMAIN_NAME, container.getSubdomain());
        containerMap.put(CONF_HOST, container.getHost());

        if (container.getNetworkBindingPort() != 0) {
            containerMap.put(CONF_NETWORK_BIND_PORT, Integer.toString(container
                    .getNetworkBindingPort()));
        }

        containerMap.put(CONF_USER, container.getUser());
        containerMap.put(CONF_PWD, container.getPassword());

        containerMap.put(CONF_JMX_RMI_PORT, Integer.toString(container.getJmxRMIConnectorPort()));

        if (container.getJoramDomainPort() != 0) {
            containerMap.put(CONF_JORAM_DOMAIN_PORT, Integer.toString(container
                    .getJoramDomainPort()));
            containerMap.put(CONF_JORAM_TCP_PORT, Integer.toString(container.getJoramTCPPort()));
        }
        if (container.getDreamTCPPort() != 0) {
            containerMap.put(CONF_DREAM_TCP_PORT, Integer.toString(container.getDreamTCPPort()));
        }

        if (ContainerConfiguration.ContainerState.STARTED.equals(container.getState())) {
            containerMap.put(CONF_STATE, STATE_STARTED);
        } else {
            containerMap.put(CONF_STATE, STATE_STOPPED);
        }

        return containerMap;
    }
}
