/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ComponentClassLoaderTest.java 154 2006-03-27 15:30:10Z ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.platform.classloader;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.List;

/**
 * @author ofabre - eBMWebsourcing
 * @author ddesjardins - eBMWebsourcing
 */
public class ComponentClassLoader extends PetalsClassLoader {

    private SharedLibrariesClassLoader sharedLibrariesClassLoader;

    /**
     * List of the sharedlibraries that must be included in the classloading
     * strategy.
     */
    private List<String> sharedLibrariesNameList;

    public ComponentClassLoader(URL[] baseUrls,
            List<String> classpathLocations, SharedLibrariesClassLoader parent,
            List<String> sharedLibrariesNameList) throws IOException {
        super(baseUrls, classpathLocations, parent);

        this.sharedLibrariesNameList = sharedLibrariesNameList;

        this.sharedLibrariesClassLoader = parent;
    }

    /**
     * If resource is not found, search in the sharedLibrariesClassLoader, by
     * specifying the sharedLibraries to use.
     * 
     * @see java.lang.ClassLoader#getResource(java.lang.String)
     */
    @Override
    public URL getResource(String resourceName) {
        URL url = null;
        // parent first
        if (isParentFirst()) {
            url = sharedLibrariesClassLoader.getResource(resourceName,
                    sharedLibrariesNameList);
            if (url == null) {
                // Try to search in the parent classpath
                url = super.getResource(resourceName);
            }
        } else {
            // self first
            url = super.getResource(resourceName);
            if (url == null) {
                // Try to search in its own classpath
                url = sharedLibrariesClassLoader.getResource(resourceName,
                        sharedLibrariesNameList);
            }
        }
        return url;
    }

    /**
     * If resource is not found, search in the sharedLibrariesClassLoader, by
     * specifying the sharedLibraries to use.
     * 
     * @see java.lang.ClassLoader#getResourceAsStream(java.lang.String)
     */
    @Override
    public InputStream getResourceAsStream(String resourceName) {
        InputStream inputStream = null;
        // parent first
        if (isParentFirst()) {
            inputStream = sharedLibrariesClassLoader.getResourceAsStream(
                    resourceName, sharedLibrariesNameList);
            if (inputStream == null) {
                // Try to search in the parent classpath
                inputStream = super.getResourceAsStream(resourceName);
            }
        } else {
            // self first
            inputStream = super.getResourceAsStream(resourceName);
            if (inputStream == null) {
                // Try to search in its own classpath
                inputStream = sharedLibrariesClassLoader.getResourceAsStream(
                        resourceName, sharedLibrariesNameList);
            }
        }
        return inputStream;
    }

    /**
     * If class is not found, search in the sharedLibrariesClassLoader, by
     * specifying the sharedLibraries to use.
     * 
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    @Override
    protected synchronized Class<?> loadClass(String className, boolean resolve)
            throws ClassNotFoundException {
        Class<?> clazz = null;
        // search if the class is already loaded
        clazz = findLoadedClass(className);

        // class not loaded, load it
        if (clazz == null) {
            // parent first
            if (useParentFirst(className, true)) {
                try {
                    clazz = sharedLibrariesClassLoader.loadClass(className,
                            resolve, sharedLibrariesNameList);
                } catch (ClassNotFoundException cnfe) {
                    clazz = findClass(className);
                }
            } else {
                // self first
                try {
                    clazz = findClass(className);
                } catch (Throwable cnfe) {
                    // The class may be already loaded from a parent class
                    // loader
                    if (isIsolated()) {
                        throw new ClassNotFoundException(cnfe.getMessage());
                    } else {
                        clazz = sharedLibrariesClassLoader.loadClass(className,
                                resolve, sharedLibrariesNameList);
                    }
                }
            }
        }
        if (resolve) {
            resolveClass(clazz);
        }
        return clazz;
    }

}
