/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramConnection.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.transport.platform.joram;

import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import javax.jbi.messaging.MessageExchange;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.DeliveryMode;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Queue;
import javax.jms.Session;
import javax.naming.Context;
import javax.naming.NamingException;

import org.objectweb.joram.client.jms.Destination;
import org.objectweb.joram.client.jms.admin.AdminException;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.platform.TransportProtocol;
import org.ow2.petals.transport.platform.TransportProtocolListener;
import org.ow2.petals.transport.util.JoramControl;
import org.ow2.petals.transport.util.TransporterUtil;
import org.ow2.petals.util.LoggingUtil;

/**
 * Manage the JMS connection to the JORAM server
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author rnaudin - EBM WebSourcing
 * 
 */
public class JoramConnection {

    /**
     * An inner class uses for handling JMS session in a pool
     */
    class JMSSessionHandler {

        public JMSSessionHandler() {
            super();
        }

        /**
         * The JMS session
         */
        public Session session;

        /**
         * The JMS message producer
         */
        public MessageProducer producer;

        /**
         * The anonym JMS message consumer
         */
        public MessageListener messageListener;
    }

    /**
     * Size of the send <code>Session</code> pool
     */
    protected static final int SEND_POOLSIZE = 30;

    /**
     * Size of the receive <code>Session</code> pool
     */
    protected static final int RECEIVE_POOLSIZE = 10;

    /**
     * s Prefix for the topic names
     */
    protected static final String QUEUE_NAME = "JoramTransporter-";

    /**
     * Message Selector expression for the <code>SYNC_MODE</code> consumers
     */
    protected static final String SYNCMODE_MSGSELECTOR = TransporterUtil.SYNC_MODE + " = TRUE";

    /**
     * Message Selector expression for the 'normal' consumers
     */
    public static final String NORMAL_MSGSELECTOR = TransporterUtil.SYNC_MODE + " IS NULL";

    protected final Context joramContext;

    protected Connection connection;

    /**
     * Pool of <code>Session</code> used to produce JMS messages
     */
    protected final BlockingQueue<JMSSessionHandler> sendSessionHandlersPool;

    /**
     * Map of pool of <code>Session</code> used to consume JMS messages
     */
    protected final Map<String, Vector<JMSSessionHandler>> asyncReceiveSessionHandlersMap;

    /**
     * Map of pool of <code>Session</code> used to consume JMS messages
     */
    protected final Map<String, Vector<JMSSessionHandler>> syncReceiveSessionHandlersMap;

    /**
     * Flag which indicates if the JMS connection is started or not
     */
    protected boolean connectionStarted = false;

    /**
     * A serializer instance
     */
    private final ObjectSerializer serializer = new ObjectSerializer();

    /**
     * The local container configuration
     */
    protected final ContainerConfiguration containerConfiguration;

    private final TransportProtocol joramTransporter;

    /**
     * The Joram Control singleton
     */
    private final JoramControl joramControl;

    protected final LoggingUtil log;

    /**
     * Creates a new instance of {@link JoramConnection}
     * 
     * @param containerInformation
     * @param rootContext
     * @param log
     */
    public JoramConnection(TransportProtocol joramTransporter,
            final ContainerConfiguration containerConfiguration, final Context joramContext,
            final LoggingUtil log) {
        this.containerConfiguration = containerConfiguration;
        this.joramContext = joramContext;
        this.log = log;
        this.joramTransporter = joramTransporter;
        this.asyncReceiveSessionHandlersMap = new HashMap<String, Vector<JMSSessionHandler>>();
        this.syncReceiveSessionHandlersMap = new HashMap<String, Vector<JMSSessionHandler>>();
        this.sendSessionHandlersPool = new ArrayBlockingQueue<JMSSessionHandler>(SEND_POOLSIZE);
        this.joramControl = JoramControl.getInstance();
    }

    /**
     * Build the logical queue name according to the container and component
     * names.
     * 
     * @param containerName
     * @param componentName
     * 
     * @return the logical name of the Queue
     */
    public static final String getQueueName(String containerName, String componentName) {

        StringBuilder QueueNameBuilder = new StringBuilder(QUEUE_NAME);
        QueueNameBuilder.append(containerName).append("-").append(componentName);

        return QueueNameBuilder.toString();
    }

    /**
     * Retrieve the component name from a logical queue name according to the
     * given container name.
     * 
     * @param queueName
     *            the logical queue name
     * @param containerName
     *            a container name
     * 
     * @return the component name or null if the given queue name isn't the name
     *         of a component queue or doesn't match the given container name
     */
    public static final String getComponentName(String queueName, String containerName) {
        String componentName = null;
        String queueNamePrefix = QUEUE_NAME + containerName;
        if (queueName.contains(queueNamePrefix)) {
            componentName = queueName.substring(queueNamePrefix.length() + 1);
        }
        return componentName;
    }

    /**
     * Start the JMS connection and initialize the pool of <code>Sessions</code>
     * used to send messages.
     * 
     * @param exceptionListener
     *            The exception listener of this connection. Used to receive
     *            asynchronous JMS errors. <b>MUST</b> be non null
     * @throws JMSException
     * @throws ConnectException
     */
    protected void start(final ExceptionListener exceptionListener) throws JMSException,
            ConnectException {
        this.log.start();
        if (exceptionListener == null) {
            throw new NullPointerException("Parameter exceptionListener must not be null");
        }

        ConnectionFactory connectionFactory = this.joramControl.getConnectionFactory();

        this.log.debug("create JMS Connection with user/pwd:"
                + this.containerConfiguration.getUser() + "/"
                + this.containerConfiguration.getPassword());

        // create the connection with the user/pwd, or the default one if no
        // user is specified
        if (this.containerConfiguration.getUser() != null) {
            this.connection = connectionFactory.createConnection(this.containerConfiguration
                    .getUser(), this.containerConfiguration.getPassword());
        } else {
            this.connection = connectionFactory.createConnection();
        }

        this.connection.setExceptionListener(exceptionListener);

        // create the sessions pool to send JMS messages
        for (int i = 0; i < SEND_POOLSIZE; i++) {
            JMSSessionHandler jmsSessionHandler = new JMSSessionHandler();
            jmsSessionHandler.session = this.connection.createSession(false,
                    Session.AUTO_ACKNOWLEDGE);
            jmsSessionHandler.producer = jmsSessionHandler.session.createProducer(null);
            this.sendSessionHandlersPool.add(jmsSessionHandler);
        }

        this.connection.start();
        this.connectionStarted = true;

        this.log.end();
    }

    /**
     * Interrupt all pending received exchanges. Close all related the JMS
     * sessions. Wait that all ongoing send finished or that the timeout is
     * reached. Close all related the JMS sessions.
     * 
     * @throws JMSException
     *             Can not close the JMS connection or session
     * @throws NamingException
     *             Can not close the JNDI context access
     */
    protected void stop() throws JMSException, NamingException {
        this.log.start();
        this.connectionStarted = false;

        for (Vector<JMSSessionHandler> receiveJMSSessionHandlers : this.asyncReceiveSessionHandlersMap
                .values()) {

            // Close all the sessions of the async receive pool.
            // For synchronism issue,
            // the session are closed within the Message Listener.
            for (JMSSessionHandler jmsSessionHandler : receiveJMSSessionHandlers) {
                ((AsyncExchangeMessageListener) jmsSessionHandler.messageListener).terminate();
            }
        }

        for (Vector<JMSSessionHandler> receiveJMSSessionHandlers : this.syncReceiveSessionHandlersMap
                .values()) {
            // Close all the sessions of the sync receive pool.
            for (JMSSessionHandler jmsSessionHandler : receiveJMSSessionHandlers) {
                jmsSessionHandler.session.close();
            }

            // Wait that all the ongoing recept threads awake the matched issuer
            // thread,
            // to avoid the lost of received exchange
            ((SyncExchangeMessageListener) receiveJMSSessionHandlers.get(0).messageListener)
                    .terminate();
        }

        // wait that all ongoing send finished,
        // or that the timeout is reached
        int lapsTime = 0;
        try {
            while (lapsTime < Transporter.STOP_TIMEOUT) {
                lapsTime += 1000;

                if (this.sendSessionHandlersPool.size() < SEND_POOLSIZE) {
                    Thread.sleep(1000);
                } else {
                    break;
                }
            }
        } catch (InterruptedException e) {
            // do nothing
        }

        if (lapsTime >= Transporter.STOP_TIMEOUT) {
            this.log.warning("The timeout for stopping the Transporter has been reached."
                    + "Some messages to send may have been lost.");
        }

        this.connection.setExceptionListener(null);
        this.connection.stop();
        this.connection.close();
        this.log.end();
    }

    /**
     * Create and register a JMS queue to receive the component messages. Create
     * the devices to receive normal exchanges and synchronous exchanges. If a
     * device creation fails, the queue is unregistered from the JNDI server.
     * 
     * @param componentName
     *            The name of the component. <b>MUST</b> be non null
     * @param dispatcher
     *            The Transporter dispatcher. <b>MUST</b> be non null
     * @param asyncExchangesQueue
     *            The Queue where to store the received pending exchanges.
     *            <b>MUST</b> be non null
     * 
     * @throws AdminException
     * @throws NamingException
     * @throws ConnectException
     * @throws AdminException
     * @throws UnknownHostException
     * @throws JMSException
     */
    protected void createQueueDevice(final String componentName,
            final TransportProtocolListener dispatcher,
            final ArrayBlockingQueue<MessageExchange> asyncExchangesQueue) throws ConnectException,
            NamingException, AdminException, UnknownHostException, JMSException {
        this.log.start();
        if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        } else if (dispatcher == null) {
            throw new NullPointerException("Parameter dispatcher must not be null");
        } else if (asyncExchangesQueue == null) {
            throw new NullPointerException("Parameter asyncExchangesQueue must not be null");
        }

        // create and bind the related queue
        Queue localQueue = this.bindQueue(componentName);

        // retrieve the logical queue Name
        String queueName = getQueueName(this.containerConfiguration.getName(), componentName);

        try {
            // create the sync exchange device
            this.createSyncExchangeDevice(componentName, queueName, localQueue, dispatcher);

            // create the async exchange device
            this.createAsyncExchangeDevice(componentName, queueName, localQueue, dispatcher,
                    asyncExchangesQueue);
        } catch (JMSException e) {
            this.unbindQueue(queueName);
            throw e;
        }

        this.log.end();
    }

    /**
     * Try to retrieve the bound queue and create it if doesn't exist. Bind or
     * rebind it if necessary. Connection to the Joram AdminModule is done
     * internally.
     * 
     * @param componentName
     *            The name of the component. <b>MUST</b> be non null
     * 
     * @return the JMS queue. <b>MUST</b> be non null
     * 
     * @throws NamingException
     * @throws ConnectException
     * @throws AdminException
     * @throws UnknownHostException
     */
    protected Queue bindQueue(final String componentName) throws NamingException, ConnectException,
            AdminException, UnknownHostException {
        this.log.start();

        Queue queue = null;
        String queueName = getQueueName(this.containerConfiguration.getName(), componentName);

        boolean alreadyBound = true;
        try {
            queue = (Queue) this.joramContext.lookup(queueName);
            if (queue == null) {
                alreadyBound = false;
            }
        } catch (NamingException e) {
            alreadyBound = false;
        }

        if (alreadyBound) {

            // check if the topic exits really on the server; he may not due to
            // a previous crash
            if (!this.joramControl.isDestinationExist((Destination) queue, Integer
                    .parseInt(this.containerConfiguration.getName()))) {
                // create and rebind the topic
                this.log.warning("The component '" + componentName
                        + "' may have lost messages due to a previous crash.");
                queue = this.joramControl.createQueue(queueName);
                this.joramContext.rebind(queueName, queue);
            }

        } else {
            // create and bind a new topic
            queue = this.joramControl.createQueue(queueName);
            this.joramContext.bind(queueName, queue);
        }

        if (queue == null) {
            throw new NullPointerException("The returned value must be non null");
        }

        this.log.end();
        return queue;
    }

    /**
     * Delete and unbind the queue if the queue is defined on the JNDI server.
     * Connection to the JOrAM AdminModule is done internally.
     * 
     * @param componentName
     *            The name of the component
     * 
     * @throws NamingException
     * @throws ConnectException
     * @throws UnknownHostException
     * @throws AdminException
     * @throws JMSException
     */
    protected void unbindQueue(String queueName) throws NamingException, ConnectException,
            UnknownHostException, AdminException, JMSException {
        this.log.start();

        boolean alreadyBound = true;
        if (this.joramContext.lookup(queueName) == null) {
            alreadyBound = false;
        }

        if (alreadyBound) {
            org.objectweb.joram.client.jms.Queue queue = (org.objectweb.joram.client.jms.Queue) this.joramContext
                    .lookup(queueName);
            // unbind the queue
            this.joramContext.unbind(queueName);

            // delete the existing queue
                queue.delete();
        }
        this.log.end();
    }

    /**
     * Send the MessageExchange to the destination of the given component. The
     * exchange is serialized into a jms Message.
     * 
     * @param containerName
     *            The name of the destination container
     * @param componentName
     *            The name of the destination component
     * @param messageExchange
     *            The message exchange to serialized to a
     *            <code>JMS.ObjectMessage</code>
     * @param timeToLive
     *            The duration of availability of the JMS message in JORAM
     * @param isSyncMode
     *            SYNC_MODE is activated or not
     * @param isPersistent
     *            <code>True</code> if the JMS transfer must be persistent
     * @throws TransportException
     */
    protected void sendToDestination(String containerName, String componentName,
            MessageExchangeImpl messageExchange, long timeToLive, boolean isSyncMode,
            boolean isPersistent) throws TransportException {
        this.log.start();

        JMSSessionHandler sessionHandler = null;

        try {
            if (this.connectionStarted == false) {
                Thread.currentThread().interrupt();
                throw new InterruptedException("The Transporter is stopped");
            }

            // find the topic destination in jndi
            Queue remoteQueue = this.findQueue(containerName, componentName);

            // retrieve a session - block until a session is available
            sessionHandler = this.sendSessionHandlersPool.take();
            // the transpoter may be stopped at this point
            if (this.connectionStarted == false) {
                Thread.currentThread().interrupt();
                throw new InterruptedException("The Transporter is stopped");
            }

            // serialise MessageExchange object to a JMS Object Message
            Message msg = this.serializer.jbi2jms(messageExchange, sessionHandler.session);

            // set the SYNC_MODE JMS property to true
            if (isSyncMode) {
                msg.setBooleanProperty(TransporterUtil.SYNC_MODE, true);
            }

            if (isPersistent) {
                sessionHandler.producer.send(remoteQueue, msg, DeliveryMode.PERSISTENT, 0,
                        timeToLive);
            } else {
                sessionHandler.producer.send(remoteQueue, msg, DeliveryMode.NON_PERSISTENT, 0,
                        timeToLive);
            }
        } catch (InterruptedException e) {
            throw new TransportException(e);
        } catch (JMSException e) {
            throw new TransportException("Can not send the message to the JMS destination", e);
        } finally {
            // release the JMS session handler
            if (sessionHandler != null) {
                this.sendSessionHandlersPool.offer(sessionHandler);
            }
        }
        this.log.end();
    }

    /**
     * Remove the MessageConsumers bound to the component Queue. The queue is
     * removed from the Joram configuration and from the Jndi server.
     * 
     * @param componentName
     * 
     * @throws JMSException
     * @throws NamingException
     * @throws ConnectException
     * @throws UnknownHostException
     * @throws AdminException
     * @throws TransportException
     */
    protected void deleteQueueDevice(String componentName) throws JMSException, NamingException,
            ConnectException, UnknownHostException, AdminException, TransportException {
        this.log.start();

        String queueName = getQueueName(this.containerConfiguration.getName(), componentName);

        Vector<JMSSessionHandler> receiveJMSSessionHandlers = this.asyncReceiveSessionHandlersMap
                .remove(queueName);

        // Close all the sessions of the async receive pool.
        // For synchronism issue,
        // the session are closed within the Message Listener.
        for (JMSSessionHandler jmsSessionHandler : receiveJMSSessionHandlers) {
            ((AsyncExchangeMessageListener) jmsSessionHandler.messageListener).terminate();
        }

        receiveJMSSessionHandlers = this.syncReceiveSessionHandlersMap.remove(queueName);
        // Close all the sessions of the receive pool.
        for (JMSSessionHandler jmsSessionHandler : receiveJMSSessionHandlers) {
            jmsSessionHandler.session.close();
        }

        // Wait that all the ongoing recept threads awake the matched issuer
        // thread,
        // to avoid the lost of received exchange.
        ((SyncExchangeMessageListener) receiveJMSSessionHandlers.get(0).messageListener)
                .terminate();

        this.unbindQueue(queueName);

        this.log.end();
    }

    /**
     * Find in the JNDI context the registered Queue associated to the specified
     * containerName and componentName. The Queue can be either local or remote.
     * 
     * @param componentName
     * @param containerName
     * 
     * @return the Queue related to the containerName
     * 
     * @throws TransportException
     *             Queue not found, or jndi access exception
     */
    protected Queue findQueue(String containerName, String componentName) throws TransportException {
        this.log.start();

        Queue remoteQueue = null;

        try {
            remoteQueue = (Queue) this.joramContext.lookup(getQueueName(containerName,
                    componentName));
        } catch (NamingException e) {
            throw new TransportException(
                    "Can not find the Queue associated to the destination component '"
                            + componentName + "'", e);
        }

        this.log.end();
        return remoteQueue;
    }

    /**
     * Create the device for handling synchronous exchanges. Create a JMS
     * session, a JMS consumer with the synchronous message selector. Set a
     * {@link SyncExchangeMessageListener} on it. If the creation fails, the
     * created JMS sessions are closed.
     * 
     * @param componentName
     *            The name of the component. <b>MUST</b> be non null
     * @param queueName
     *            The name of the JMS queue. <b>MUST</b> be non null
     * @param localQueue
     *            The JMS queue. <b>MUST</b> be non null
     * @param dispatcher
     *            The Transporter dispatcher. <b>MUST</b> be non null
     * 
     * @throws JMSException
     */
    private void createSyncExchangeDevice(final String componentName, final String queueName,
            final Queue localQueue, final TransportProtocolListener dispatcher) throws JMSException {
        Vector<JMSSessionHandler> receiveSessionHandlers = null;
        SyncExchangeMessageListener messageListener = new SyncExchangeMessageListener(this.log,
                dispatcher, componentName);

        try {
            // create the pool of JMS session handler
            receiveSessionHandlers = new Vector<JMSSessionHandler>(RECEIVE_POOLSIZE);

            for (int i = 0; i < RECEIVE_POOLSIZE; i++) {
                JMSSessionHandler sessionHandler = new JMSSessionHandler();
                sessionHandler.session = this.connection.createSession(false,
                        Session.AUTO_ACKNOWLEDGE);
                MessageConsumer consumer = sessionHandler.session.createConsumer(localQueue,
                        SYNCMODE_MSGSELECTOR);
                consumer.setMessageListener(messageListener);
                sessionHandler.messageListener = messageListener;
                receiveSessionHandlers.add(sessionHandler);
            }
            this.syncReceiveSessionHandlersMap.put(queueName, receiveSessionHandlers);
        } catch (JMSException exception) {
            // clean the device
            this.log.warning("Fail to create the synchronous exchanges device", exception);
            try {
                if (receiveSessionHandlers != null) {
                    for (JMSSessionHandler sessionHandler : receiveSessionHandlers) {
                        sessionHandler.session.close();
                    }
                }
            } catch (JMSException e) {
                // do nothing
            }
            throw exception;
        }
    }

    /**
     * Create the device for handling normal exchanges. Create a pool of JMS
     * Sessions. Each session creates a Consumer with an async message selector.
     * Set a {@link AsyncExchangeMessageListener} on them. If the creation
     * fails, the created JMS sessions are closed.
     * 
     * @param componentName
     *            The name of the component. <b>MUST</b> be non null
     * @param queueName
     *            The name of the queue. <b>MUST</b> be non null
     * @param localQueue
     *            The queue. <b>MUST</b> be non null
     * @param dispatcher
     *            The Transporter dispatcher. <b>MUST</b> be non null
     * @param asyncExchangesQueue
     *            The Queue where to store the received pending exchanges.
     *            <b>MUST</b> be non null
     * 
     * @throws JMSException
     */
    private void createAsyncExchangeDevice(final String componentName, final String queueName,
            final Queue localQueue, final TransportProtocolListener dispatcher,
            final ArrayBlockingQueue<MessageExchange> asyncExchangesQueue) throws JMSException {
        Vector<JMSSessionHandler> receiveSessionHandlers = null;

        try {
            // create the pool of JMS session handler
            receiveSessionHandlers = new Vector<JMSSessionHandler>(RECEIVE_POOLSIZE);

            for (int i = 0; i < RECEIVE_POOLSIZE; i++) {
                JMSSessionHandler sessionHandler = new JMSSessionHandler();
                sessionHandler.session = this.connection.createSession(false,
                        Session.CLIENT_ACKNOWLEDGE);
                MessageConsumer consumer = sessionHandler.session.createConsumer(localQueue,
                        NORMAL_MSGSELECTOR);
                AsyncExchangeMessageListener messageListener = new AsyncExchangeMessageListener(
                        this.log, joramTransporter, dispatcher, asyncExchangesQueue, componentName,
                        sessionHandler.session);
                consumer.setMessageListener(messageListener);
                sessionHandler.messageListener = messageListener;
                receiveSessionHandlers.add(sessionHandler);
            }
            this.asyncReceiveSessionHandlersMap.put(queueName, receiveSessionHandlers);
        } catch (JMSException exception) {
            // clean the device
            this.log.warning("Fail to create the asynchronous exchanges device", exception);
            try {
                if (receiveSessionHandlers != null) {
                    for (JMSSessionHandler sessionHandler : receiveSessionHandlers) {
                        sessionHandler.session.close();
                    }
                }
            } catch (JMSException e) {
                // do nothing
            }
            throw exception;
        }
    }
}
