/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramTransporter.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.transport.platform.joram;

import java.io.File;
import java.io.IOException;
import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Set;
import java.util.concurrent.ArrayBlockingQueue;

import javax.jbi.messaging.MessageExchange;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.kernel.configuration.ConfigurationException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.kernel.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.transport.ProtocolMonitor;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.platform.TransportProtocol;
import org.ow2.petals.transport.platform.TransportProtocolListener;
import org.ow2.petals.transport.util.JoramControl;
import org.ow2.petals.transport.util.JoramTopology;
import org.ow2.petals.transport.util.TransporterUtil;
import org.ow2.petals.util.JNDIUtil;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.SystemUtil;

/**
 * Implementation of the <code>TransporterProtocol</code> with JORAM.<br>
 * In dynamic mode, at the starting of the Transporter, the JORAM configuration
 * file is created according to the PEtALS topology, then the JORAM server is
 * started. For each component installed, the <code>JoramTransporter</code>
 * creates and registers a <code>Queue</code> in the distributed JNDI server
 * with the following name:
 * <em>JoramConnection-[containerName]-[componentName]</em><br>
 * The name of a container is the <em>ID</em> specified in the
 * <em>conf/server.properties</em> file. <br>
 * Similarly, at each uninstallation of component, the <code>Queue</code> is
 * deleted and unregistered. <br>
 * This <code>JoramTransporter</code> uses an <code>ObjectSerializer</code>
 * to create <em>jms.ObjectMessage</em> to send via JMS. So, the
 * <code>MessageExchange</code> to send is serialized.
 * 
 * @version $Rev: 477 $ $Date: 2006-05-29 17:18:07 +0200 (lun., 29 mai 2006) $
 * @since Petals 1.0
 * @author alouis
 * @author rnaudin
 * @author garciap
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = TransportProtocol.class))
public class JoramTransporter implements TransportProtocol {

    /**
     * Name used to identify the JORAM transporter when the "transporterToUse"
     * property is set.
     */
    public static final String TRANSPORTER_NAME = "joram";

    /**
     * MONoLOG component logger
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * The logging utility
     */
    protected LoggingUtil log;

    /**
     * The Joram Connection instance
     */
    private JoramConnection joramConnection;

    /**
     * The local container configuration
     */
    protected ContainerConfiguration containerConfiguration;

    /**
     * The domain configuration
     */
    protected DomainConfiguration domainConfiguration;

    /**
     * JOrAM controller singleton.
     */
    private JoramControl joramControl;

    /**
     * JOrAM controller singleton.
     */
    private JoramMonitoring joramMonitoring;

    /**
     * JOrAM topology singleton
     */
    private JoramTopology joramTopology;

    /**
     * Flag indicating if the container has been shut down
     */
    private boolean isShutDown = false;

    /**
     * The JORAM directory
     */
    private static File joramDirectory;

    /**
     * Map of blocking queues for the received exchanges waiting to be consumed
     * by the dispatcher.
     */
    private Hashtable<String, ArrayBlockingQueue<MessageExchange>> asyncExchangesQueuesMap;

    /**
     * The Topology service fractal component
     */
    @Requires(name = "topology", signature = org.ow2.petals.communication.topology.TopologyService.class)
    protected TopologyService topologyService;

    /**
     * The JNDI service fractal component
     */
    @Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The dispatcher fractal component
     */
    @Requires(name = "dispatcher", signature = org.ow2.petals.transport.platform.TransportProtocolListener.class)
    protected TransportProtocolListener dispatcher;

    /**
     * The Fractal component
     */
    @Service
    Component joramTransporterComponent;

    /**
     * The <code>Transporter</code> creates a topic for the given component
     * and bind it to the JNDI server. It creates a subscriber to poll the
     * 'normal' messages and a listener to receiving the response messages when
     * synchronious send has been done.
     * 
     * @throws TransportException
     */
    public void addDestination(final String componentName) throws TransportException {
        if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        }
        log.start("Add destination : " + componentName);

        ArrayBlockingQueue<MessageExchange> asyncExchangesQueue = new ArrayBlockingQueue<MessageExchange>(
                JoramConnection.RECEIVE_POOLSIZE);

        asyncExchangesQueuesMap.put(componentName, asyncExchangesQueue);

        try {
            joramConnection.createQueueDevice(componentName, dispatcher, asyncExchangesQueue);
        } catch (Exception e) {
            String msg = "Cannot add the destination for the component '" + componentName + "'";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
        log.end();
    }

    /**
     * The <code>Transporter</code> removes the <code>Sesssions</code> bound
     * to the component related Topic. Then it deletes the Topic and unbind it
     * from the JNDI server.
     */
    public void removeDestination(final String componentName) throws TransportException {
        if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        }
        log.start("Remove destination : " + componentName);

        try {
            joramConnection.deleteQueueDevice(componentName);
        } catch (Exception e) {
            String msg = "Cannot remove the destination for the component '" + componentName + "'";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }
        log.end();
    }

    /**
     * Send the message to the specified distant or local container, by
     * serializing the exchange with a {@link ObjectSerializer}. We retrieve
     * the targeted {@link javax.jms.Queue} with the name of the container and
     * the component, transform the {@link MessageExchangeImpl} into a
     * {@link javax.jms.Message}, and send this message to the related
     * {@link javax.jms.Queue}. <br/> <br/> If the
     * "org.objectweb.petals.transport.nonpersistent" exchange property is set
     * to true, the message is not persisted by the JORAM layer, but it can be
     * lost in case of problem.
     * 
     * @see org.ow2.petals.transport.Transporter#send(org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl,
     *      java.lang.String, java.lang.String, long)
     */

    public void send(MessageExchangeImpl messageExchange, final String componentName,
            final String containerName, final long timeToLive) throws TransportException {

        if (messageExchange == null) {
            throw new NullPointerException("Parameter messageExchange must not be null");
        } else if (containerName == null) {
            throw new NullPointerException("Parameter containerName must not be null");
        } else if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        } else if (timeToLive < 0) {
            throw new NullPointerException("Parameter timeToLive must be positive or zero");
        }

        log.start("Send message to container : " + containerName + " / component : "
                + componentName);

        boolean persistent = true;
        if (Transporter.FAST_POLICY.equals(messageExchange
                .getProperty(Transporter.PROPERTY_TRANSPORT_QOS))) {
            persistent = false;
        }

        try {
            joramConnection.sendToDestination(containerName, componentName, messageExchange,
                    timeToLive, TransporterUtil.getSyncMode(messageExchange, true), persistent);
        } catch (TransportException e) {
            log.error(e.getMessage(), e);
            throw e;
        }
        log.end();
    }

    /**
     * Shutdown the {@code Transporter} and unregister it from the topology. The
     * queues related to the container must have been removed from the JOrAM
     * agent and from the JNDI server previously.
     * 
     * @throws TransportException
     */
    public void shutdown() throws TransportException {
        log.start();

        isShutDown = true;
        
        try {
            if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {
                joramTopology.removeServerfromTopology(containerConfiguration, topologyService
                        .getAllContainersConfiguration(ContainerState.STARTED));
            }
       
        } catch (PetalsException e) {
            throw new TransportException("Can not remove the JORAM server from the JORAM domain", e);
        }

        log.end();
    }

    /**
     * The <code>Transporter</code> polls for a received message on the
     * component related topic.
     */
    public MessageExchangeImpl receive(final String componentName) throws TransportException {
        if (componentName == null) {
            throw new NullPointerException("Parameter componentName must not be null");
        }
        log.start("Receive : " + componentName);

        ArrayBlockingQueue<MessageExchange> asyncExchangesQueue = asyncExchangesQueuesMap
                .get(componentName);

        MessageExchangeImpl exchange;
        exchange = (MessageExchangeImpl) asyncExchangesQueue.poll();

        // notify the exchange to awake the asynchronous message listener, which
        // acknowledge the JMS message.
        synchronized (exchange) {
            exchange.notify();
        }

        log.end();

        return exchange;
    }

    /**
     * Initialize the fields. In dynamic mode, build the JOrAM topology, then
     * start the JORAM server.
     * 
     * @throws TransportException
     *             Failed to start the JOrAM transporter
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws TransportException {
        log = new LoggingUtil(logger);
        log.start();

        try {
            if (joramDirectory == null) {
                joramDirectory = new File(SystemUtil.getPetalsInstallDirectory() + File.separator
                        + "work" + File.separator + JoramTransporter.TRANSPORTER_NAME);
                if (!joramDirectory.exists()) {
                    joramDirectory.mkdir();
                }
            }

            joramControl = JoramControl.getInstance();

            joramTopology = JoramTopology.getInstance();

            asyncExchangesQueuesMap = new Hashtable<String, ArrayBlockingQueue<MessageExchange>>();

            // get the local and domain configuration
            containerConfiguration = configurationService.getContainerConfiguration();
            domainConfiguration = configurationService.getDomainConfiguration();

            // build the JOrAM Topology
            buildJoramTopology();

            // start the JOrAM agent
            startJoramAgent();

            Context joramContext = createJoramJndiContext();

            TransportProtocol joramTransportProtocol = (TransportProtocol) joramTransporterComponent
                    .getFcInterface("service");

            joramConnection = new JoramConnection(joramTransportProtocol, containerConfiguration,
                    joramContext, log);

            // create an exception listener to log all asynchronous errors
            ExceptionListener exceptionListener = new ExceptionListener() {
                public void onException(JMSException exception) {
                    log.error("Asynchronious JMS exception", exception);
                    return;
                }
            };

            joramConnection.start(exceptionListener);

            // Create a Joram monitoring instance
            joramMonitoring = new JoramMonitoring(containerConfiguration);

        } catch (ConnectException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start JOrAM Transporter service", e);
            throw new TransportException(e);
        } catch (JMSException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start JOrAM Transporter service", e);
            throw new TransportException(e);
        } catch (NoSuchInterfaceException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start JOrAM Transporter service", e);
            throw new TransportException(e);
        } catch (NamingException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start JOrAM Transporter service", e);
            throw new TransportException(e);
        } catch (PetalsException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start JOrAM Transporter service", e);
            throw new TransportException(e);
        } catch (TransportException e) {
            // log the error here as Fractal layer is removing the Exception
            // stack
            log.error("Failed to start JOrAM Transporter service", e);
            throw e;
        }

        log.end();
    }

    /**
     * Close the JMS connection and stop the JORAM server. The
     * <code>Transporter</code> stay registered in the distributed JNDI
     * server.
     * 
     * @throws TransportException
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws TransportException {
        log.start();

        try {
            this.joramConnection.stop();
            if (isShutDown) {
                this.joramControl.disconnectFromAdminModule();
            }
            this.joramControl.stopAgentServer();
            // remove the JORAM directory if the container has been shut down
            if (isShutDown) {
                joramControl.cleanAgentServerWorkspace(joramDirectory, true);
            }
        } catch (Exception e) {
            String msg = "Can not stop the Joram Transporter";
            log.error(msg, e);
            throw new TransportException(msg, e);
        }

        log.end();
        logger = null;
        log = null;
    }

    /**
     * Build the JOrAM topology resource.
     * 
     * @throws TransportException
     */
    private void buildJoramTopology() throws TransportException {
        log.start();

        try {
            Topology topology = null;
            if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {
                topology = topologyService.getDynamicTopology();
            } else {
                topology = configurationService.getStaticTopology();
            }

            joramTopology.buildJoramA3ServerFile(joramDirectory, topology);

        } catch (TopologyException e) {
            throw new TransportException(e);
        } catch (IOException e) {
            throw new TransportException(e);
        } catch (ConfigurationException e) {
            throw new TransportException(e);
        } catch (PetalsException e) {
            throw new TransportException(e);
        }

        log.end();
    }

    /**
     * Start the JOrAM agent.<br>
     * First initialize the agent, then start it.
     * 
     * @throws PetalsException
     */
    private void startJoramAgent() throws PetalsException {
        log.start();

        // if JORAM is not launched, initialize and start it
        if (!joramControl.isAgentServerRunning()) {

            // initialize the JORAM agent
            joramControl.initAgentServer(joramDirectory, new Integer(containerConfiguration
                    .getName()).shortValue());

            // start the JORAM agent
            try {
                joramControl.startAgentServer();
            } catch (Exception e) {
                throw new PetalsException("Can not start JORAM agent", e);
            }
        }

        // update the others JORAM servers in dynamic mode
        if (DomainConfiguration.DomainMode.DYNAMIC.equals(domainConfiguration.getMode())) {

            Set<ContainerConfiguration> containersConfiguration = topologyService
                    .getAllContainersConfiguration(ContainerState.STARTED);
            joramTopology.addServerToTopology(domainConfiguration.getDomain(),
                    containerConfiguration, containersConfiguration);
        }

        // connect to the collocated server
        try {
            this.joramControl.connectToAdminModule(this.containerConfiguration);
        } catch (ConnectException e) {
            throw new PetalsException(e);
        } catch (NumberFormatException e) {
            throw new PetalsException(e);
        } catch (UnknownHostException e) {
            throw new PetalsException(e);
        } catch (AdminException e) {
            throw new PetalsException(e);
        }

        log.end();
    }

    private Context createJoramJndiContext() throws NamingException {
        log.start();

        Context joramContext = null;
        // get the local container from the JNDI registry
        InitialContext rootContext = jndiService.getInitialContext();
        if (JNDIUtil.isBound(rootContext, "/", TRANSPORTER_NAME)) {
            joramContext = (Context) rootContext.lookup(TRANSPORTER_NAME);
        } else {
            joramContext = rootContext.createSubcontext(TRANSPORTER_NAME);
        }

        log.end();
        return joramContext;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.platform.TransportProtocol#getProtocolMonitor()
     */
    public ProtocolMonitor getProtocolMonitor() {
        return joramMonitoring;
    }
}
