/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): pierre.garcia@inrialpes.fr
 */

package org.ow2.petals.transport.util;

import java.util.HashMap;
import java.util.Map;

import javax.jbi.JBIException;

import org.objectweb.dream.message.ChunkAlreadyExistException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.InvalidReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.ow2.petals.PetalsException;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.platform.dream.MessageExchangeChunk;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.util.LoggingUtil;

public class DreamUtil {

    /**
     * Name of the chunk including destination
     */
    protected static final String DESTINATION_CHUNK_NAME = "destination";

    /**
     * Message Manager. Used here only to manage chunk in a message
     */
    protected MessageManagerType messageManagerItf;

    /**
     * Factory used to create Export Identifier Chunk
     */
    protected ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

    /**
     * Factory used to create, delete, copy... MessageExchangeChunk
     */
    ChunkFactoryReference<MessageExchangeChunk> chunkFactory;

    /**
     * Inherited Logger
     */
    LoggingUtil log;

    /**
     * The Topology server
     */
    protected TopologyService topologyService;

    /**
     * Mapping between container name and ipExportChunk
     */
    protected Map<String, ExportIdentifierChunk> exportIdChunks;

    public DreamUtil(MessageManagerType messageManager, LoggingUtil log,
            TopologyService topologyService) {
        this.log = log;
        this.log.start();

        this.messageManagerItf = messageManager;
        // create the chunkFactory
        this.exportIdChunkFactory = this.messageManagerItf
                .getChunkFactory(ExportIdentifierChunk.class);
        this.chunkFactory = this.messageManagerItf.getChunkFactory(MessageExchangeChunk.class);
        this.topologyService = topologyService;

        this.exportIdChunks = new HashMap<String, ExportIdentifierChunk>();

        this.log.end();
    }

    /**
     * Add a chunk including the destination of the message
     * 
     * @param message
     * @param host
     * @param port
     * @return
     * @throws TransportException
     */
    public Message addChunkDestination(Message message, String containerName)
            throws TransportException {
        this.log.start();

        ExportIdentifierChunk eic = this.exportIdChunks.get(containerName);

        if (eic == null) {
            ContainerConfiguration containerConfiguration = null;
            try {
                containerConfiguration = this.topologyService
                        .getContainerConfiguration(containerName);
            } catch (PetalsException e) {
                throw new TransportException("Cannot get the information for the container "
                        + containerName);
            }
            eic = this.createChunkDestination(containerConfiguration);
        }
        eic = this.messageManagerItf.cloneChunk(eic);
        this.log.debug("Chunk: " + eic);
        this.messageManagerItf.addOrReplaceChunk(message, DESTINATION_CHUNK_NAME, eic);

        this.log.end();
        return message;
    }

    /**
     * 
     * 
     * @param containerName
     * @return
     * @throws TransportException
     */
    protected ExportIdentifierChunk createChunkDestination(
            ContainerConfiguration containerConfiguration) throws TransportException {
        this.log.start();

        ExportIdentifierChunk eic = null;
        String host = containerConfiguration.getHost();
        int port = containerConfiguration.getDreamTCPPort();
        IPExportIdentifier exportIdentifier = new IPExportIdentifier(host, port);
        eic = this.messageManagerItf.createChunk(this.exportIdChunkFactory);
        eic.setExportIdentifier(exportIdentifier);

        this.exportIdChunks.put(containerConfiguration.getName(), eic);

        this.log.end("Export identifier (" + host + ", " + port + ")");
        return eic;
    }

    /**
     * Transform a jbi message into a message formatted for a specific router.
     * 
     * @param jbi
     *            the message to transform
     * @param parameter
     *            map of specific needed objects required by the transporter.
     * @return the message formatted for the transporter
     * @throws JBIException
     *             thrown if an error occurs during serialization
     */
    public MessageExchangeImpl dream2jbi(Message dreamMessage) throws JBIException {
        this.log.start();

        try {
            MessageExchangeChunk mec = (MessageExchangeChunk) this.messageManagerItf.getChunk(
                    dreamMessage, MessageExchangeChunk.DEFAULT_NAME);
            if (mec == null) {
                throw new JBIException("MessageExchangeChunk not found !");
            } else {
                this.log.end();
                return mec.getMessageExchange();
            }
        } catch (InvalidReference ir) {
            throw new JBIException(ir);
        }
    }

    /**
     * Transform a transporter format message into a jbi message.
     * 
     * @param messageFromTransporter
     *            message with a specific format related to the transporter
     * @param stuffs
     *            map of specific needed.
     * @return a jbi {@link MessageExchangeImpl message}
     * @throws TransportException
     *             thrown if an error occurs during serialization
     */
    public Message jbi2dream(MessageExchangeImpl jbi) throws TransportException {
        this.log.start();

        Message dreamMessage = this.messageManagerItf.createMessage();
        try {
            MessageExchangeChunk mec = this.messageManagerItf.createChunk(this.chunkFactory);
            mec.setMessageExchange(jbi);
            this.messageManagerItf.addChunk(dreamMessage, MessageExchangeChunk.DEFAULT_NAME, mec);
            this.log.end();
            return dreamMessage;
        } catch (InvalidReference ir) {
            throw new TransportException(ir);
        } catch (ChunkAlreadyExistException caee) {
            throw new TransportException(caee);
        }
    }
}
